from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *

from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_interstellar_lines", "plot_interstellar_lines"]

subroutine_name = 'interstellar_lines'


# def plot_identify_stellar_lines(config_in)

def compute_interstellar_lines(config_in):

    night_dict = from_config_get_nights(config_in)

    instrument_dict = from_config_get_instrument(config_in)

    interstellar_lines = from_config_get_interstellar_lines(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    if not interstellar_lines:
        return

    for night in night_dict:
        print()
        print("compute_interstellar_lines                 Night: ", night)

        try:
            interstellar = load_from_cpickle('interstellar_lines_processed', config_in['output'], night)
            skip_lineselection = True
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        instrument = night_dict[night]['instrument']

        processed = {
            'subroutine': subroutine_name,
            'line_rebin': {},
            'line_shift': {}
        }

        interstellar = {
            'subroutine': subroutine_name,
        }

        for obs in lists['observations']:
            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            interstellar[obs] = {
                'wave_BRF': shift_wavelength_array(input_data[obs]['wave'], observational_pams[obs]['rv_shift_ORF2BRF']),
                'correction':  np.ones(np.shape(input_data[obs]['wave']))
            }

            """ for plotting purpose only"""
            processed[obs]['flux'] = input_data[obs]['e2ds'] / calib_data['blaze'] / input_data[obs]['step']
            processed[obs]['flux_err'] = input_data[obs]['e2ds_err'] / calib_data['blaze'] / input_data[obs][
                'step']

        for line_name, line in interstellar_lines.items():

            processed[line_name] = {
                'line_rebin': {},
                'poly_coeff': {},
                'normalized': {},
                'line_shift': {
                    'selected_points': []
                }
            }

            processed[line_name]['min_wave'] = max(shared_data['coadd']['wavelength_range'][0], line[0] - line[2]*2)
            processed[line_name]['max_wave'] = min(shared_data['coadd']['wavelength_range'][1], line[0] + line[2]*2)

            processed[line_name]['wave'] = np.arange(processed[line_name]['min_wave'],
                                                     processed[line_name]['max_wave'],
                                                     instrument_dict[instrument]['wavelength_step'])

            processed[line_name]['size'] = np.size(processed[line_name]['wave'], axis=0)
            processed[line_name]['step'] = np.ones(processed[line_name]['size'])\
                * instrument_dict[instrument]['wavelength_step']

            processed[line_name]['correction'] = np.ones(processed[line_name]['size'])

            for obs in lists['observations']:

                preserve_flux = input_data[obs].get('absolute_flux', True)

                """ shifting a chunk of the spectra to the Solar System Barycenter reference """
                processed[line_name]['line_rebin'][obs] = \
                    rebin_2d_to_1d(input_data[obs]['wave'],
                                   input_data[obs]['step'],
                                   input_data[obs]['e2ds'],
                                   calib_data['blaze'],
                                   processed[line_name]['wave'],
                                   processed[line_name]['step'],
                                   preserve_flux=preserve_flux,
                                   rv_shift=observational_pams[obs]['rv_shift_ORF2BRF'])

                argmin_sel = np.argmin(processed[line_name]['line_rebin'][obs][5:-5]) + 5
                wave_sel = processed[line_name]['wave'][argmin_sel]

                processed[line_name]['line_shift']['selected_points'].append(wave_sel)

            processed[line_name]['line_shift']['position'] = np.median(
                processed[line_name]['line_shift']['selected_points'])
            processed[line_name]['line_shift']['delta_lambda'] = processed[line_name]['line_shift']['position'] - line[0]

            try:
                if interstellar_lines['automatic'] == True:
                    interstellar[line_name] = processed[line_name]['line_shift']['position']
                else:
                    interstellar[line_name] = line[0]
            except:
                interstellar[line_name] = line[0]

            """ selection of spectral range for continuum normalization and interstellar line modelling"""

            processed[line_name]['wavelength_selection'] = \
                (np.abs(processed[line_name]['wave'] - interstellar[line_name]) < line[1])

            processed[line_name]['continuum_selection'] = \
                (~processed[line_name]['wavelength_selection']) \
                & (np.abs(processed[line_name]['wave'] - interstellar[line_name]) < line[2])

            processed[line_name]['interstellar_selection'] = \
                (processed[line_name]['wavelength_selection'] | processed[line_name]['continuum_selection'])

            spline_wave_points = []
            spline_norm_points = []

            # TODO
            #! 1) Rescale by median each observation and collect all the value
            #! 2) Perform a continuum normalization on the collect values, with
            #! iterative sigma-clipping
            #! 3) Perform a spline / gaussian fit of the spectral line

            for obs in lists['telluric']:

                # sel1 = (np.abs(processed[line_name]['wave'] - interstellar[line_name]) < line[1])
                # sel2 = (~sel1) & (np.abs(processed[line_name]['wave'] - interstellar[line_name]) < line[2])
                # sel3 = (sel1 | sel2)

                """ normalization around the interstellar line """
                processed[line_name]['poly_coeff'][obs] = \
                    np.polyfit(processed[line_name]['wave'][processed[line_name]['continuum_selection']],
                               processed[line_name]['line_rebin'][obs][processed[line_name]['continuum_selection']],
                               2)

                processed[line_name]['normalized'][obs] = \
                    processed[line_name]['line_rebin'][obs][processed[line_name]['interstellar_selection']] \
                    / np.polyval(processed[line_name]['poly_coeff'][obs],
                                 processed[line_name]['wave'][processed[line_name]['interstellar_selection']])

                spline_wave_points.extend(processed[line_name]['wave'][processed[line_name]['interstellar_selection']])
                spline_norm_points.extend(processed[line_name]['normalized'][obs])

            """ sorting the array to avoid problems with the spline function"""
            spline_sorting_index = np.argsort(spline_wave_points)
            spline_wave_points = np.asarray(spline_wave_points)[spline_sorting_index]
            spline_norm_points = np.asarray(spline_norm_points)[spline_sorting_index]

            processed[line_name]['spline_eval'], \
                processed[line_name]['spline_coeff'], \
                processed[line_name]['spline_knots'] = \
                compute_spline(spline_wave_points, spline_norm_points, 0.08, knot_order=3)

            processed[line_name]['correction'][processed[line_name]['wavelength_selection']] = \
                sci_int.splev(processed[line_name]['wave'][processed[line_name]['wavelength_selection']],
                              processed[line_name]['spline_coeff'])

            for obs in lists['observations']:

                interstellar[obs]['wavelength_selection'] = \
                    (np.abs(interstellar[obs]['wave_BRF']-interstellar[line_name]) < line[1])
                interstellar[obs]['continuum_selection'] = \
                    (~interstellar[obs]['wavelength_selection']) \
                    & (np.abs(interstellar[obs]['wave_BRF']-interstellar[line_name]) < line[2])
                interstellar[obs]['interstellar_selection'] = \
                    (interstellar[obs]['wavelength_selection'] | interstellar[obs]['continuum_selection'])

                interstellar[obs]['correction'][interstellar[obs]['wavelength_selection']] = \
                    sci_int.splev(interstellar[obs]['wave_BRF'][interstellar[obs]['wavelength_selection']],
                                  processed[line_name]['spline_coeff'])

        save_to_cpickle('interstellar_lines_processed', processed, config_in['output'], night)
        save_to_cpickle('interstellar_lines', interstellar, config_in['output'], night)


def plot_interstellar_lines(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    interstellar_lines = from_config_get_interstellar_lines(config_in)

    if not interstellar_lines:
        return

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        print("plot_interstellar_lines                    Night: ", night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('interstellar_lines_processed', config_in['output'], night)
            interstellar = load_from_cpickle('interstellar_lines', config_in['output'], night)
        except:
            print()
            print('No interstellar correction, no plots')
            continue

        colors_properties, colors_plot, colors_scatter = make_color_array_matplotlib3(lists, observational_pams)

        # fig = plt.figure(figsize=(12, 6))
        # gs = GridSpec(2, 2, width_ratios=[50, 1])
        #
        # ax1 = plt.subplot(gs[0, 0])
        # ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        # cbax1 = plt.subplot(gs[:, 1])
        fig, gs, cbax1, ax1, ax2 = grid_2plot()

        for i, obs in enumerate(lists['observations']):

            """rescaling"""
            processed[obs]['flux_rescaling'], processed[obs]['flux_rescaled'], processed[obs]['flux_rescaled_err'] = \
                perform_rescaling(interstellar[obs]['wave_BRF'],
                                  processed[obs]['flux'],
                                  processed[obs]['flux_err'],
                                  observational_pams['wavelength_rescaling'])

            if i == 0:
                ax1.scatter(interstellar[obs]['wave_BRF'], processed[obs]['flux_rescaled'],
                            c=colors_scatter['mBJD'][obs], s=1, alpha=0.5, label='observations (BRF)')
            else:
                ax1.scatter(interstellar[obs]['wave_BRF'], processed[obs]['flux_rescaled'],
                            c=colors_scatter['mBJD'][obs], s=1, alpha=0.5)

            # ax1.plot(interstellar['wave'], interstellar['correction'], c='black')

            ax2.scatter(interstellar[obs]['wave_BRF'], processed[obs]['flux_rescaled']/interstellar[obs]['correction'],
                        c=colors_scatter['mBJD'][obs], s=1, alpha=0.5)

        for line_name, line in interstellar_lines.items():

            ax1.axvline(interstellar[line_name]-line[1], c='b')
            ax1.axvline(interstellar[line_name]+line[1], c='b')
            ax1.axvline(interstellar[line_name]-line[2], c='g')
            ax1.axvline(interstellar[line_name]+line[2], c='g')

            ax2.axvline(interstellar[line_name]-line[1], c='b')
            ax2.axvline(interstellar[line_name]+line[1], c='b')
            ax2.axvline(interstellar[line_name]-line[2], c='g')
            ax2.axvline(interstellar[line_name]+line[2], c='g')

            # ax1.plot(processed[line_name]['wave'], processed[line_name]['flux_rescaled'], c='b')

            try:
                wave_min = min(wave_min, interstellar[line_name])
                wave_max = max(wave_max, interstellar[line_name])
                range_max = max(range_max, line[2])
            except:
                wave_min = interstellar[line_name]
                wave_max = interstellar[line_name]
                range_max = line[2]

        ax1.set_title('Night: {0:s} \n Input spectra'.format(night))
        ax1.set_xlim(wave_min-2*range_max, wave_max+2*range_max)
        ax1.legend(loc=1)

        ax2.set_title('After interstellar line correction')
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        # fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()



import numpy as np
from scipy.optimize import curve_fit


def interpolate1d_grid_nocheck(val, arr_1, arr_2):
    # using enumerate() + next() to find index of
    # first element just greater than 0.6
    res = next(x for x, v in enumerate(arr_1)  if v > val)
    interp_out =  (val-arr_1[res-1])/(arr_1[res]-arr_1[res-1])*(arr_2[res,:]-arr_2[res-1,:])  + arr_2[res-1,:]
    return interp_out

def interpolate2d_grid_nocheck(val, arr_1, arr_2):
    # using enumerate() + next() to find index of
    # first element just greater than 0.6
    res = next(x for x, v in enumerate(arr_1)  if v > val)
    interp_out =  (val-arr_1[res-1])/(arr_1[res]-arr_1[res-1])*(arr_2[res,:,:]-arr_2[res-1,:,:])  + arr_2[res-1,:,:]
    return interp_out

def first_derivative(x_arr, y_arr):
    n_points = len(x_arr)
    derivative = np.zeros(n_points, dtype=np.double)

    derivative[1:-1] = (y_arr[2:] - y_arr[:-2]) / (x_arr[2:] - x_arr[:-2])
    derivative[0] = derivative[1]
    derivative[-1] = derivative[-2]

    return derivative



"""from classes.kepler_exo import *

# Mass of the star HD189733 (in Solar masses)
#Ms = 0.823
Ms = 1.148
# Mass of the planet (in Solar masses)
#Mp = 1.138 / 1.047348644e3
Mp = 0.69 / 1.047348644e3

K1 = kepler_K1(Mp,Ms,3.52474854657,86.59,0.0082)
print K1


## update
"""

import matplotlib.pyplot as plt
import numpy as np
from SLOPpy.subroutines.constants import *
import argparse
from scipy.optimize import fsolve
from SLOPpy.subroutines.kepler_exo import *

def get_mass(M_star2, M_star1, Period, K1, e0):
    # M_star1, M_star2 in solar masses
    # P in days -> Period is converted in seconds in the routine
    # inclination assumed to be 90 degrees
    # Gravitational constant is given in m^3 kg^-1 s^-2
    # output in m/s
    output = K1 - (2. * np.pi * G_grav * M_sun / 86400.0) ** (1.0 / 3.0) * (1.000 / np.sqrt(1.0 - e0 ** 2.0)) * (
                                                                                                                    Period) ** (
                                                                                                                    -1.0 / 3.0) * (
                      M_star2 * (M_star1 + M_star2) ** (-2.0 / 3.0))
    return output


parser = argparse.ArgumentParser(prog='planetary_velocity.py', description='Compute the expected semi-amplitude of the planet')
parser.add_argument('star_mass', type=float, nargs=1, help='Stellar mass [solar]')
parser.add_argument('plan_mass', type=float, nargs=1, help='Planet mass [Jupiter/Earth/K units, default Jupiter]')
parser.add_argument('period', type=float, nargs=1, help='Planetary period [days')
parser.add_argument('inclination', type=float, nargs=1, help='Planetary inclination [degrees]')
parser.add_argument('eccentricity', type=float, nargs=1, help='Planetary eccentricity [pure]')
parser.add_argument('-e', type=float, nargs='?', default=False, const=True, help='Planetary mass in Earth units')
parser.add_argument('-k', type=float, nargs='?', default=False, const=True, help='Planetary mass in m/s')
args = parser.parse_args()


star_mass = args.star_mass[0]
P = args.period[0]
i = args.inclination[0]
e = args.eccentricity[0]
if args.e and args.k:
    print('Either -k or -e, not both!')
    quit()

planet_mass = args.plan_mass[0]
if args.e:
    planet_mass *= Mears
elif args.k:
    x0 = Mjups
    K_input = planet_mass
    planet_mass = fsolve(get_mass, x0, args=(star_mass, P, K_input, e))
else:
    planet_mass *= Mjups

print(kepler_K1(planet_mass, star_mass, P, i, e))


#sampler = args.sample[0]
#file_conf = args.config_file[0]



# -*- coding: utf-8 -*-
#
# PyORBIT documentation build configuration file, created by
# sphinx-quickstart on Sat Dec 16 17:20:15 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import matplotlib
matplotlib.use('agg')

sys.path.insert(0, os.path.abspath('../'))
import SLOPpy


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['myst_parser',
    "sphinx.ext.autosummary",
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    "sphinx.ext.viewcode",
    'sphinx.ext.napoleon']

myst_enable_extensions = ["dollarmath", "colon_fence"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
#source_suffix = ['.rst', '.md']
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
}
#source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'SLOPpy'
copyright = u'2018-2022, Daniela Sicilia, Luca Malavolta'
author = u'Daniela Sicilia, Luca Malavolta'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '.'.join(SLOPpy.__version__.split('.')[:-1])
# The full version, including alpha/beta/rc tags.
release = SLOPpy.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'
#html_logo = "_static/PyORBIT_logo_transp.png"
html_title = "SLOPpy"

html_copy_source = True
html_show_sourcelink = True
html_sourcelink_suffix = ""

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/LucaMalavolta/SLOPpy",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_issues_button": True,
    #"home_page_in_toc": True,
    "show_navbar_depth": 3,
    "logo_only": True,
    #"show_related": True
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#

#html_sidebars = {
#    '**': ['sidebar-logo.html','search-field.html',
#        'globaltoc.html',
#        'relations.html',  # needs 'show_related': True theme option to display
##        #'searchbox.html',
#    ]
#}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'SLOPpy_doc'



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["compute_telluric_airmass_berv_observerRF",
           "plot_telluric_airmass_berv_observerRF",
           "compute_telluric_airmass_observerRF",
           "plot_telluric_airmass_observerRF",
           "compute_telluric_airmass_reference_observerRF",
           "plot_telluric_airmass_reference_observerRF",
           "compute_telluric_airmass_berv_reference_observerRF",
           "plot_telluric_airmass_berv_reference_observerRF"
           ]


def compute_telluric_airmass_berv_observerRF(config_in):

    compute_telluric_observerRF(config_in,
                                n_iterations=1,
                                use_berv=True,
                                use_reference_airmass=False,
                                subroutine_name='telluric_airmass_berv_observerRF')


def compute_telluric_airmass_observerRF(config_in):

    compute_telluric_observerRF(config_in,
                                n_iterations=1,
                                use_berv=False,
                                use_reference_airmass=False,
                                subroutine_name='telluric_airmass_observerRF')


def compute_telluric_airmass_reference_observerRF(config_in):

    compute_telluric_observerRF(config_in,
                                n_iterations=1,
                                use_berv=False,
                                use_reference_airmass=True,
                                subroutine_name='telluric_airmass_reference_observerRF')

def compute_telluric_airmass_berv_reference_observerRF(config_in):
    compute_telluric_observerRF(config_in,
                                n_iterations=1,
                                use_berv=True,
                                use_reference_airmass=True,
                                subroutine_name='telluric_airmass_berv_reference_observerRF')


def plot_telluric_airmass_berv_observerRF(config_in, night_input):
    """ Alias to simplify the configuration file"""
    plot_telluric_airmass_observerRF(config_in, night_input)


def plot_telluric_airmass_reference_observerRF(config_in, night_input):
    """ Alias to simplify the configuration file"""
    plot_telluric_airmass_observerRF(config_in, night_input)


def plot_telluric_airmass_berv_reference_observerRF(config_in, night_input):
    """ Alias to simplify the configuration file"""
    plot_telluric_airmass_observerRF(config_in, night_input)


def compute_telluric_observerRF(config_in, **kwargs):

    night_dict = from_config_get_nights(config_in)

    for night in night_dict:

        print()
        print("compute_telluric_airmass_observerRF        Night: ", night)

        try:
            telluric = load_from_cpickle('telluric', config_in['output'], night)
            continue
        except:
            print("No telluric correction file found, computing now ")
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        """ Retrieving the observations"""
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'], use_telluric=False)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': kwargs['subroutine_name'],
            'n_orders': 0,
            'n_pixels': 0
        }

        telluric = {
            'subroutine': kwargs['subroutine_name'],
            'reference_frame': 'observer'
        }

        # There must be a more elegant way to do this, but I'm, not aware of it
        for obs in lists['observations']:

            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            """ for plotting purpose only"""
            processed[obs]['wave'] = input_data[obs]['wave']
            processed[obs]['e2ds'] = input_data[obs]['e2ds']
            processed[obs]['e2ds_err'] = input_data[obs]['e2ds_err']

            processed[obs]['e2ds_rescaling'], processed[obs]['e2ds_rescaled'], processed[obs]['e2ds_rescaled_err'] = \
                perform_rescaling(input_data[obs]['wave'],
                                  input_data[obs]['e2ds'],
                                  input_data[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            if processed['n_orders'] == 0:
                processed['n_orders'] = input_data[obs]['orders']
                processed['n_pixels'] = input_data[obs]['wave_size']

        """ Reference airmass for iterative correction of airmass"""
        if kwargs['use_reference_airmass']:
            airmass_temp = np.zeros(lists['n_transit_in'])
            for n_obs, obs in enumerate(lists['transit_in']):
                # This is to ensure that airmass, berv and rvc are associated to the correct spectra
                airmass_temp[n_obs] = input_data[obs]['AIRMASS']
            processed['airmass_ref'] = np.average(airmass_temp)
        else:
            processed['airmass_ref'] = 0.000

        for obs in lists['observations']:
            processed[obs]['e2ds_precorrected'] = processed[obs]['e2ds_rescaled'][:]
            processed[obs]['e2ds_precorrected_err'] = input_data[obs]['e2ds_err'] / processed[obs]['e2ds_rescaling']

        for niter in range(0, kwargs['n_iterations']):
            if kwargs['n_iterations'] > 1:
                print("NITER:            ", niter)

            for obs in lists['telluric']:
                processed[obs]['logI'] = np.log(processed[obs]['e2ds_precorrected'])
                processed[obs]['logI_err'] = processed[obs]['e2ds_precorrected_err']/processed[obs]['e2ds_precorrected']

            processed['telluric'] = {}

            abs_slope = np.ones([processed['n_orders'], processed['n_pixels']], dtype=np.double)
            line_shift = np.ones([processed['n_orders'], processed['n_pixels']], dtype=np.double)
            zero_point = np.ones([processed['n_orders'], processed['n_pixels']], dtype=np.double)
            pearson_r = np.zeros([processed['n_orders'], processed['n_pixels']], dtype=np.double)
            pearson_p = np.zeros([processed['n_orders'], processed['n_pixels']], dtype=np.double)

            airmass = np.zeros(lists['n_tellurics'], dtype=np.double)
            berv = np.zeros(lists['n_tellurics'], dtype=np.double)
            rvc = np.zeros(lists['n_tellurics'], dtype=np.double)

            for n_obs, obs in enumerate(lists['telluric']):
                # This is to ensure that airmass, berv and rvc are associated to the correct spectra
                processed['telluric'][obs] = {'n_obs': n_obs}
                airmass[n_obs] = input_data[obs]['AIRMASS']
                berv[n_obs] = input_data[obs]['BERV']
                rvc[n_obs] = input_data[obs]['RVC']

            for order in range(0, processed['n_orders']):

                logi_array = np.empty([lists['n_tellurics'], processed['n_pixels']], dtype=np.double)
                sigi_array = np.empty([lists['n_tellurics'], processed['n_pixels']], dtype=np.double)

                for obs in lists['telluric']:
                    n_obs = processed['telluric'][obs]['n_obs']
                    logi_array[n_obs, :] = processed[obs]['logI'][order, :]
                    sigi_array[n_obs, :] = processed[obs]['logI_err'][order, :]

                """ The user has the option to select between different approaches to
                    extract the telluric absorption spectrum
                    To-Do: move this section to a subroutine for cythonization"""

                if kwargs['use_berv']:
                    if observational_pams['linear_fit_method'] == 'linear_curve_fit':

                        abs_slope[order, :], line_shift[order, :], zero_point[order, :] = \
                            berv_linear_curve_fit_modified(airmass, berv, logi_array, sigi_array, processed['n_pixels'])

                    else:
                        abs_slope[order, :], line_shift[order, :], zero_point[order, :] = \
                            berv_linear_lstsq(airmass, berv, logi_array)

                else:
                    if observational_pams['linear_fit_method'] == 'linear_curve_fit':

                        abs_slope[order, :], zero_point[order, :] = \
                            airmass_linear_curve_fit(airmass, logi_array, sigi_array, processed['n_pixels'])

                        #abs_slope[order, :], zero_point[order, :] = \
                        #    airmass_linear_curve_fit_ransac(airmass, logi_array, sigi_array, processed['n_pixels'])

                        #obs_ref = lists['observations'][0]
                        #plt.plot(processed[obs_ref]['wave'][order,:], processed[obs_ref]['e2ds_rescaled'][order,:])
                        #for iii in range(0, processed['n_pixels']):
                        #    if iii < 3700 or iii > 3720: continue
                        #    plt.axvline(processed[obs_ref]['wave'][order,iii])
                        #plt.show()
                        #
                        #ik=0
                        #air_arr = np.arange(1.2, 2.5, 0.1)
                        #for iii in range(0, processed['n_pixels']):
                        #
                        #    if iii < 3700 or iii > 3720: continue
                        #    plt.errorbar(airmass, logi_array[:, iii]+ik, yerr=sigi_array[:, iii], fmt='o')
                        #    print(np.exp(abs_slope[order, iii]))
                        #    plt.plot(air_arr, air_arr*abs_slope[order, iii] + zero_point[order, iii]+ik)
                        #    ik -= 0.20
                        #plt.show()
                    else:

                        abs_slope[order, :], zero_point[order, :] = \
                            airmass_linear_lstsq(airmass, logi_array)

                plt.show()
                """ Saving the outcome to dictionary """
                processed['telluric']['order_'+repr(order)] = {'logi_array': logi_array, 'sigi_array': sigi_array}

            if kwargs.get('use_template', False):
                telluric_template_data = np.genfromtxt(night_dict[night]['telluric_template'])

                spectrum_noairmass = np.exp(abs_slope)

                obs_reference = lists['observations'][0]
                telluric['template'] = {
                    'input':{
                        'wave': telluric_template_data[:, 0],
                        'flux': telluric_template_data[:, 1],
                        'ferr': telluric_template_data[:, 2],
                        'step': telluric_template_data[:, 3]
                    },
                    'rebinned':{
                        'wave': input_data[obs_reference]['wave'],
                        'step': input_data[obs_reference]['step']
                    }
                }

                telluric['template']['rebinned']['flux'] = \
                    rebin_1d_to_2d(telluric['template']['input']['wave'],
                                   telluric['template']['input']['step'],
                                   telluric['template']['input']['flux'],
                                   telluric['template']['rebinned']['wave'],
                                   telluric['template']['rebinned']['step'],
                                   preserve_flux=False)

                telluric['template']['rebinned']['ferr'] = \
                    rebin_1d_to_2d(telluric['template']['input']['wave'],
                                   telluric['template']['input']['step'],
                                   telluric['template']['input']['ferr'],
                                   telluric['template']['rebinned']['wave'],
                                   telluric['template']['rebinned']['step'],
                                   preserve_flux=False,
                                   is_error=True)

                plt.plot(telluric['template']['input']['wave'], telluric['template']['input']['flux'], zorder=1, c='C0')
                plt.scatter(telluric['template']['rebinned']['wave'], telluric['template']['rebinned']['flux'],zorder=2, s=2)
                plt.scatter(telluric['template']['rebinned']['wave'],spectrum_noairmass, alpha=0.5, s=1, zorder=3)

                factor_list = []
                slope_list = []

                for order in range(0, processed['n_orders']):
                    fit_selection = (telluric['template']['rebinned']['flux'][order, :] < 1.0)
                    # Check if there are telluric lines in this wavelength range
                    if np.sum(fit_selection) > 30:
                        #telluric_factor, telluric_slope, success_flag = find_telluric_rescaling_factor(
                        #    spectrum_noairmass[order, :],
                        #    telluric['template']['rebinned']['flux'][order, :]
                        #)
                        telluric_factor, telluric_slope, success_flag = find_telluric_rescaling_factor_2steps(
                            spectrum_noairmass[order, :],
                            telluric['template']['rebinned']['flux'][order, :]
                        )

                        if success_flag:
                            factor_list.extend([telluric_factor])
                            slope_list.extend([telluric_slope])

                if len(factor_list)>1:
                    telluric_slope = np.median(slope_list)
                    telluric_factor = np.median(factor_list)
                elif len(factor_list) == 1:
                    telluric_slope = slope_list[0]
                    telluric_factor = factor_list[0]
                else:
                    telluric_slope = 0.00
                    telluric_factor = 0.00

                print('   telluric factor: {0:7f} (correction slope: {1:7f}'.format(telluric_factor,telluric_slope))
                print()
                #print(telluric_factor, success_flag)
                #quit()

                processed['telluric']['spectrum_noairmass'] = \
                    (telluric['template']['rebinned']['flux']-1.)*telluric_factor + 1.0

                plt.plot(telluric['template']['input']['wave'],
                         (telluric['template']['input']['flux']-1.)*telluric_factor + 1.0, zorder=1, c='C1')

                plt.show()


            else:
                processed['telluric']['spectrum_noairmass'] = np.exp(abs_slope)

            for obs in lists['observations']:
                """ Correction of telluric lines for the average airmass value, following Wyttenbach et al. 2015 """
                processed[obs]['e2ds_corrected'] = processed[obs]['e2ds_precorrected'] / \
                                                        np.power(processed['telluric']['spectrum_noairmass'],
                                                                 input_data[obs]['AIRMASS'] -
                                                                 processed['airmass_ref'])
                processed[obs]['e2ds_corrected_err'] = processed[obs]['e2ds_precorrected_err'] / \
                                                        np.power(processed['telluric']['spectrum_noairmass'],
                                                                 input_data[obs]['AIRMASS'] -
                                                                 processed['airmass_ref'])

        for obs in lists['observations']:
            # Correction of telluric lines

            telluric[obs] = {}

            telluric[obs]['spectrum_noairmass'] = processed['telluric']['spectrum_noairmass']

            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']
            telluric[obs]['airmass_ref'] = processed['airmass_ref']
            telluric[obs]['null'] = telluric[obs]['spectrum_noairmass'] < 0.001
            telluric[obs]['spectrum_noairmass'][ telluric[obs]['null']] = 1.0

            telluric[obs]['spectrum'] = np.power(processed['telluric']['spectrum_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            telluric[obs]['spline_noairmass'] = np.ones([input_data[obs]['n_orders'],
                                                         input_data[obs]['n_pixels']],
                                                        dtype=np.double)

            for order in range(0, processed['n_orders']):
                telluric[obs]['spline_noairmass'][order, :], _, _ = \
                    compute_spline(input_data[obs]['wave'][order, :],
                                   telluric[obs]['spectrum_noairmass'][order, :],
                                   0.05)

            telluric[obs]['spline'] = np.power(telluric[obs]['spline_noairmass'],
                                       input_data[obs]['AIRMASS'] - processed['airmass_ref'])

            telluric[obs]['telluric_corrected'] = processed[obs]['e2ds_corrected']
            telluric[obs]['telluric_corrected_err'] = processed[obs]['e2ds_corrected_err']

        save_to_cpickle('telluric', telluric, config_in['output'], night)
        save_to_cpickle('telluric_processed', processed, config_in['output'], night)

        print()
        print("Night ", night, " completed")


def plot_telluric_airmass_observerRF(config_in, night_input=''):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        print("plot_telluric_airmass_observerRF           Night: ", night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('telluric_processed', config_in['output'], night)
            telluric = load_from_cpickle('telluric', config_in['output'], night)
        except:
            print()
            print("No telluric correction, no plots")
            continue

        colors, cmap, line_colors = make_color_array(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        lift_spectrum = 0.25

        for i, obs in enumerate(lists['observations']):
            color_array = cmap(i / len(lists['observations']))

            _, e2ds_rescaled , _ = \
                perform_rescaling(processed[obs]['wave'],
                                  processed[obs]['e2ds'],
                                  processed[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            e2ds_rescaled_corrected_spectrum = e2ds_rescaled / telluric[obs]['spectrum']
            e2ds_rescaled_corrected_spline = e2ds_rescaled / telluric[obs]['spline']

            for order in range(0, processed[obs]['n_orders']):

                if order == 0 and i==0:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5, label='uncorrected')
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array), label='corrected')
                else:
                    ax1.plot(processed[obs]['wave'][order, :],
                                e2ds_rescaled[order, :],
                                c=color_array, lw=1, alpha=0.5)
                    ax1.scatter(processed[obs]['wave'][order, :],
                                e2ds_rescaled_corrected_spectrum[order, :],
                                s=1, c=np.atleast_2d(color_array))

                #ax1.plot(processed[obs]['wave'][order, :],
                #            e2ds_rescaled[order, :]+lift_spectrum,
                #            c=color_array, lw=1, alpha=0.5)
                #ax1.scatter(processed[obs]['wave'][order, :],
                #            e2ds_rescaled_corrected_spline[order, :]+lift_spectrum,
                #            s=1, c=np.atleast_2d(color_array))

                ax2.plot(processed[obs]['wave'][order, :],
                         telluric[obs]['spectrum'][order, :],
                         c=color_array)
                ax2.axhline(1.00, c='k')

                #ax2.plot(processed[obs]['wave'][order, :],
                #         telluric[obs]['spline'][order, :]+lift_spectrum,
                #         c=color_array)
                #ax2.axhline(1.00+lift_spectrum, c='k')

        #ax2.plot(input_data['coadd']['wave'],telluric['stellarRF']['spline_eval']+0.1,c='k')
        #ax2.scatter(input_data['coadd']['wave'],telluric['stellarRF']['spectrum']+0.1,c='r', s=2)

        ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        try:
            instrument = night_dict[night]['instrument']
            comparison_file = config_in['instruments'][instrument]['telluric_comparison']
            comparison_data = np.genfromtxt(comparison_file, skip_header=1)
            if comparison_data[0,0]<1000.0:
                nm2Ang = 10.
            else:
                nm2Ang = 1.
            ax1.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
            ax2.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
        except:
            pass


        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


import numpy as np
from SLOPpy.subroutines.rebin_subroutines import *


def clv_rm_correction_factor_computation(clv_rm_modelling,  wave, step, rv_shift, obs):
    ancillary = {}

    ancillary['norm_convolved_shifted'] = \
        rebin_1d_to_1d(clv_rm_modelling['common']['wave'],
                       clv_rm_modelling['common']['step'],
                       clv_rm_modelling['common']['norm_convolved'],
                       wave,
                       step,
                       rv_shift=rv_shift,
                       preserve_flux=False)

    ancillary['stellar_spectra_convolved_shifted'] = \
        rebin_1d_to_1d(clv_rm_modelling['common']['wave'],
                       clv_rm_modelling['common']['step'],
                       clv_rm_modelling[obs]['stellar_spectra_convolved'],
                       wave,
                       step,
                       rv_shift=rv_shift,
                       preserve_flux=False)

    wavelength_exclusion = \
        (wave <= clv_rm_modelling['common']['wave'][0] + 1) | \
        (wave >= clv_rm_modelling['common']['wave'][-1] - 1)

    wavelength_selection = \
        (wave > clv_rm_modelling['common']['wave'][0] + 1) & \
        (wave > clv_rm_modelling['common']['wave'][-1] - 1)

    ancillary['norm_convolved_shifted'][wavelength_exclusion] = \
        np.amax(ancillary['norm_convolved_shifted'][wavelength_selection])
    ancillary['stellar_spectra_convolved_shifted'][wavelength_exclusion] = \
        np.amax(ancillary['stellar_spectra_convolved_shifted'][wavelength_selection])

    ancillary['correction'] = ancillary['stellar_spectra_convolved_shifted'] / ancillary['norm_convolved_shifted']

    return ancillary['correction'], ancillary



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.math_functions import *
from SLOPpy.transmission_spectrum_preparation import compute_transmission_spectrum_preparation
from scipy.signal import savgol_filter

__all__ = ['compute_transmission_spectrum',
           'plot_transmission_spectrum']


subroutine_name = 'transmission_spectrum'
sampler_name = 'emcee'


def compute_transmission_spectrum(config_in, lines_label, reference='planetRF', night_input='', preparation_only=False, pca_iteration=-1):

    results_list_default = ['user',
                            'mcmc_night_MED',
                            'mcmc_night_MAP',
                            'mcmc_global_MED',
                            'mcmc_global_MAP']
    # compute_transmission_spectrum_preparation(config_in)

    night_dict = from_config_get_nights(config_in)
    ### transmission_dict = from_config_get_transmission(config_in)
    pca_parameters = from_config_get_pca_parameters(config_in)

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]

    norm_dict = lines_dict.get('normalization', {})
    norm_pams = {}
    norm_pams['normalize_transmission'] = norm_dict.get('normalize_transmission', True)
    norm_pams['normalization_model'] = norm_dict.get('normalization_model', 'polynomial')

    """ Normalization parameters for polynomial model"""
    norm_pams['model_poly_degree'] = norm_dict.get('model_poly_degree', 2)
    norm_pams['spectra_poly_degree'] = norm_dict.get('spectra_poly_degree', 2)
    norm_pams['lower_threshold'] = norm_dict.get('lower_threshold', 0.950)
    norm_pams['percentile_selection'] = norm_dict.get('percentile_selection', 10)

    """ Normalization parameters using Savitzky-Golay filter"""
    norm_pams['window_length'] = norm_dict.get('window_length', 101)
    norm_pams['polyorder'] = norm_dict.get('polyorder', 3)
    norm_pams['mode'] = norm_dict.get('mode', 'nearest')
    norm_pams['cval'] = norm_dict.get('cval', 1.0)

    shared_data = load_from_cpickle('shared', config_in['output'])

    clv_rm_correction = lines_dict.get('clv_rm_correction', True)

    """ Using the line-specific range to define the transmission spectrum region """

    shared_selection = (shared_data['coadd']['wave'] >= lines_dict['range'][0]) \
        & (shared_data['coadd']['wave'] < lines_dict['range'][1])
    binned_selection = (shared_data['binned']['wave'] >= lines_dict['range'][0]) \
        & (shared_data['binned']['wave'] < lines_dict['range'][1])

    transmission_template = {
        'subroutine': subroutine_name,
        'range': lines_dict['range'],
        'wave': shared_data['coadd']['wave'][shared_selection],
        'step': shared_data['coadd']['step'][shared_selection],
        'size': int(np.sum(shared_selection)),
        'binned_wave': shared_data['binned']['wave'][binned_selection],
        'binned_step': shared_data['binned']['step'][binned_selection],
        'binned_size': int(np.sum(binned_selection))
    }

    for night in night_dict:

        print()
        print("Running {0:45s} for  {1:20s}   Night:{2:15s}  ".format(subroutine_name, lines_label, night))

        preparation_input = load_from_cpickle('transmission_preparation', config_in['output'], night)

        if preparation_input.get('pca_output', False):
            if pca_iteration >= 0:
                it_string = str(pca_iteration).zfill(2)
            else:
                it_string = str(pca_parameters.get('ref_iteration', 3)).zfill(2)
            preparation = preparation_input[it_string]
        else:
            preparation = preparation_input
            it_string = ''

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        results_list = results_list_default.copy()

        binned_mcmc_night = check_existence_cpickle(
            'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string)
        binned_mcmc_global = check_existence_cpickle(
            'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label, it_string=it_string)

        mcmc_night = check_existence_cpickle(
            'transmission_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string=it_string)
        mcmc_global = check_existence_cpickle(
            'transmission_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label, it_string=it_string)

        if mcmc_night and mcmc_global:
            mcmc_results_night = load_from_cpickle(
                'transmission_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string=it_string)

            mcmc_results_global = load_from_cpickle(
                'transmission_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label, it_string=it_string)

            print('  Observational parameters from MCMC fit of unbinned data and configuration file')

        elif binned_mcmc_night and binned_mcmc_global:
            mcmc_results_night = load_from_cpickle(
                'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string=it_string)

            mcmc_results_global = load_from_cpickle(
                'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label, it_string=it_string)

            print('  Observational parameters from MCMC fit of binned data and configuration file')

        else:
            print('  Observational parameters from configuration file')
            results_list = ['user']

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        if clv_rm_correction:
            try:
                clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night, lines_label)
            except (FileNotFoundError, IOError):
                clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night)

        for results_selection in results_list:

            try:
                transmission = load_from_cpickle(subroutine_name+'_'+reference + '_' +
                                                 results_selection, config_in['output'], night, lines_label, it_string=it_string)
                print("{0:45s} Night:{1:15s}   {2:s}   {3:s}   {4:s}".format(
                    subroutine_name, night, lines_label, results_selection, 'Retrieved'))
                continue
            except (FileNotFoundError, IOError):
                print("{0:45s} Night:{1:15s}   {2:s}   {3:s}   {4:s}".format(
                    subroutine_name, night, lines_label, results_selection, 'Computing'))

            transmission = transmission_template.copy()

            if len(it_string) > 0:
                transmission['pca_output'] = True
            else:
                transmission['pca_output'] = False

            print_warning = True

            for obs in lists['observations']:

                """ we start from the e2ds file, after correction for blaze and
                    division by the master-out

                Observation data:
                    wave: input_data[obs]['wave']
                    step: input_data[obs]['step']
                    flux: preparation[obs]['deblazed']
                    ferr: preparation[obs]['deblazed_err']
                """

                transmission[obs] = {}
                transmission[obs] = {
                    'BJD': input_data[obs]['BJD'],
                    'AIRMASS': input_data[obs]['AIRMASS']
                }

                """ Shift into planetary reference system is the default
                choice"""
                if results_selection == 'user':
                    planet_R_factor = observational_pams.get('Rp_factor', 1.00000)
                    if reference in ['observer', 'observerRF', 'ORF']:
                        rv_shift = 0.000
                        rv_shift_clv = -observational_pams[obs]['rv_shift_ORF2SRF']
                    elif reference in ['stellar', 'stellarRF', 'SRF']:
                        rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = 0.0000
                    else:
                        rv_shift = observational_pams[obs]['rv_shift_ORF2PRF']
                        rv_shift_clv = observational_pams[obs]['rv_shift_SRF2PRF']

                elif results_selection == 'mcmc_night_MED':
                    planet_R_factor = mcmc_results_night['results']['planet_R']
                    if reference in ['observer', 'observerRF', 'ORF']:
                        rv_shift = 0.000
                        #rv_shift_clv = mcmc_results_night['results']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = -observational_pams[obs]['rv_shift_ORF2SRF']
                    elif reference in ['stellar', 'stellarRF', 'SRF']:
                        #rv_shift = mcmc_results_night['results']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = 0.0000
                    else:
                        rv_shift = mcmc_results_night['results']['observational_pams'][obs]['rv_shift_ORF2PRF']
                        rv_shift_clv = mcmc_results_night['results']['observational_pams'][obs]['rv_shift_SRF2PRF']

                elif results_selection == 'mcmc_night_MAP':
                    planet_R_factor = mcmc_results_night['results_MAP']['planet_R']
                    if reference in ['observer', 'observerRF', 'ORF']:
                        rv_shift = 0.000
                        #rv_shift_clv = mcmc_results_night['results_MAP']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = -observational_pams[obs]['rv_shift_ORF2SRF']
                    elif reference in ['stellar', 'stellarRF', 'SRF']:
                        #rv_shift = mcmc_results_night['results_MAP']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = 0.0000
                    else:
                        rv_shift = mcmc_results_night['results_MAP']['observational_pams'][obs]['rv_shift_ORF2PRF']
                        rv_shift_clv = mcmc_results_night['results_MAP']['observational_pams'][obs]['rv_shift_SRF2PRF']

                elif results_selection == 'mcmc_global_MED':
                    planet_R_factor = mcmc_results_global['results']['planet_R']
                    if reference in ['observer', 'observerRF', 'ORF']:
                        rv_shift = 0.000
                        #rv_shift_clv = mcmc_results_global['results']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = -observational_pams[obs]['rv_shift_ORF2SRF']
                    elif reference in ['stellar', 'stellarRF', 'SRF']:
                        #rv_shift = mcmc_results_global['results']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = 0.0000
                    else:
                        rv_shift = mcmc_results_global['results']['observational_pams'][obs]['rv_shift_ORF2PRF']
                        rv_shift_clv = mcmc_results_global['results']['observational_pams'][obs]['rv_shift_SRF2PRF']

                elif results_selection == 'mcmc_global_MAP':
                    planet_R_factor = mcmc_results_global['results_MAP']['planet_R']
                    if reference in ['observer', 'observerRF', 'ORF']:
                        rv_shift = 0.000
                        #rv_shift_clv = mcmc_results_global['results_MAP']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = -observational_pams[obs]['rv_shift_ORF2SRF']
                    elif reference in ['stellar', 'stellarRF', 'SRF']:
                        #rv_shift = mcmc_results_global['results_MAP']['observational_pams'][obs]['rv_shift_ORF2SRF']
                        rv_shift = observational_pams[obs]['rv_shift_ORF2SRF']
                        rv_shift_clv = 0.0000
                    else:
                        rv_shift = mcmc_results_global['results_MAP']['observational_pams'][obs]['rv_shift_ORF2PRF']
                        rv_shift_clv = mcmc_results_global['results_MAP']['observational_pams'][obs]['rv_shift_SRF2PRF']

                """ Step 2): rebin the 2D ratio spectra to 1D """

                if transmission['pca_output']:

                    transmission[obs]['rebinned'] = \
                        rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio'],
                                    np.ones_like(calib_data['blaze']),
                                    transmission['wave'],
                                    transmission['step'],
                                    preserve_flux=False,
                                    rv_shift=rv_shift)

                    transmission[obs]['rebinned_err'] = \
                        rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio_err'],
                                    np.ones_like(calib_data['blaze']),
                                    transmission['wave'],
                                    transmission['step'],
                                    rv_shift=rv_shift,
                                    preserve_flux=False,
                                    is_error=True)
                else:
                    preserve_flux = input_data[obs].get('absolute_flux', True)

                    transmission[obs]['rebinned'] = \
                        rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio'],
                                    calib_data['blaze'],
                                    transmission['wave'],
                                    transmission['step'],
                                    preserve_flux=preserve_flux,
                                    rv_shift=rv_shift)

                    transmission[obs]['rebinned_err'] = \
                        rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio_err'],
                                    calib_data['blaze'],
                                    transmission['wave'],
                                    transmission['step'],
                                    preserve_flux=preserve_flux,
                                    rv_shift=rv_shift,
                                    is_error=True)

                    ### Small border bugfix
                    if transmission[obs]['rebinned_err'][0] ==0:
                        transmission[obs]['rebinned'][0] = transmission[obs]['rebinned'][1]
                        transmission[obs]['rebinned_err'][0] = transmission[obs]['rebinned_err'][1]

                    if transmission[obs]['rebinned_err'][-1] ==0:

                        transmission[obs]['rebinned'][-1] = transmission[obs]['rebinned'][-2]
                        transmission[obs]['rebinned_err'][-1] = transmission[obs]['rebinned_err'][-2]


                    #import matplotlib.pyplot as plt
                    #plt.scatter(transmission['wave'], transmission[obs]['corrected'])
                    #plt.plot(transmission['wave'], transmission[obs]['continuum'])
                    #plt.scatter(transmission['wave'][selection], transmission[obs]['corrected'][selection], c='r')
                    #plt.plot(input_data[obs]['wave'][0,:], preparation[obs]['ratio_err'][0,:])
                    #plt.scatter(transmission['wave'], transmission[obs]['rebinned_err'], c='b')
                    #plt.axhline(0.0000, c='C2')

                    #plt.show()
                    #quit()

                #import matplotlib.pyplot as plt
                #plt.scatter(input_data[obs]['wave'], preparation[obs]['ratio'], s=2)
                #plt.xlim(lines_dict['range'][0], lines_dict['range'][1])
                # plt.show()

                if clv_rm_correction:

                    """" CLV + RM computation in the planetary reference frame """
                    transmission[obs]['clv_model_stellarRF'] = interpolate1d_grid_nocheck(planet_R_factor,
                                                                                          clv_rm_models['common']['radius_grid'],
                                                                                          clv_rm_models[obs]['clv_rm_model_convolved_normalized'])

                    transmission[obs]['clv_model_rebinned'] = \
                        rebin_1d_to_1d(clv_rm_models['common']['wave'],
                                       clv_rm_models['common']['step'],
                                       transmission[obs]['clv_model_stellarRF'],
                                       transmission['wave'],
                                       transmission['step'],
                                       preserve_flux=False,
                                       rv_shift=rv_shift_clv)

                    #import matplotlib.pyplot as plt
                    #print(obs, planet_R_factor)
                    #plt.plot(clv_rm_models['common']['wave'], transmission[obs]['clv_model_stellarRF'], zorder=100, c='C2')
                    #plt.scatter(transmission['wave'], transmission[obs]['clv_model_rebinned'], s=2)
                    # plt.show()
                    transmission[obs]['corrected'] = transmission[obs]['rebinned'] / \
                        transmission[obs]['clv_model_rebinned']
                    transmission[obs]['corrected_err'] = transmission[obs]['rebinned_err'] / \
                        transmission[obs]['clv_model_rebinned']

                else:
                    transmission[obs]['clv_model_rebinned'] = np.ones(transmission['size'])
                    transmission[obs]['corrected'] = transmission[obs]['rebinned']
                    transmission[obs]['corrected_err'] = transmission[obs]['rebinned_err']
                    if print_warning:
                        print('   *** No CLV correction')


                if norm_pams['normalize_transmission'] and norm_pams['normalization_model'] == 'polynomial':

                    """ Continuum normalization preparatory steps:
                        1) exclusion of regions with lines of interes
                        2) exclusion of regions with stellar lines
                        3) Polynomial fit of selected regions
                        Boolean array initialized to all True values
                    """
                    transmission[obs]['line_exclusion'] = (transmission['wave'] > 0.)

                    """ Continuum normalization:
                        1) exclusion of regions with transmission lines under study, now
                        in the RF of the lines
                    """
                    for line_key, line_val in lines_dict['lines'].items():
                        transmission[obs]['line_exclusion'] = transmission[obs]['line_exclusion'] & (
                            np.abs(transmission['wave']-line_val) > 3.)

                    """ Continuum normalization:
                        2) exclusion of regions with planetary lines, taking into account the planetary RV semi-amplitude
                    """

                    if clv_rm_correction:
                        stellar_spectrum_rebinned = rebin_1d_to_1d(clv_rm_models['common']['wave'],
                                                                clv_rm_models['common']['step'],
                                                                clv_rm_models['common']['norm_convolved'],
                                                                transmission['wave'],
                                                                transmission['step'],
                                                                rv_shift=rv_shift_clv,
                                                                preserve_flux=False)

                        stellar_spectrum_derivative = first_derivative(transmission['wave'], stellar_spectrum_rebinned)

                        missing_model = (np.abs(stellar_spectrum_rebinned) < 0.0001)

                        cont_10perc = np.percentile(np.abs(stellar_spectrum_derivative), norm_pams['percentile_selection'])

                        #transmission[obs]['line_exclusion'] = transmission[obs]['line_exclusion'] \
                        #    & (np.abs(stellar_spectrum_derivative) < cont_10perc) \
                        #    & (stellar_spectrum_rebinned > norm_pams['lower_threshold'])

                        line_exclusion = transmission[obs]['line_exclusion'] \
                            & (np.abs(stellar_spectrum_derivative) < cont_10perc) \
                            & (stellar_spectrum_rebinned > norm_pams['lower_threshold'])

                        if np.sum(line_exclusion) < len(line_exclusion)/200:
                            transmission[obs]['line_exclusion'] = transmission[obs]['line_exclusion'] \
                                & ( missing_model | ((np.abs(stellar_spectrum_derivative) < cont_10perc) \
                                & (stellar_spectrum_rebinned > norm_pams['lower_threshold'])))
                        else:
                            transmission[obs]['line_exclusion'] = line_exclusion



                    elif print_warning:
                        print("   No stellar synthetic spectrum from CLV models")
                        print("   some stellar lines may be included in transmission normalization  ")
                        print_warning = False

                    """ Continuum normalization:
                        3) Polynomial fit, everything is hard coded now but personalized
                        options can be implemented easily in the yaml file
                    """

                    selection = transmission[obs]['line_exclusion'] & (
                        transmission[obs]['corrected'] > np.std(transmission[obs]['corrected']))
                    transmission[obs]['continuum_coeff'] = \
                        np.polynomial.chebyshev.chebfit(transmission['wave'][selection],
                                                        transmission[obs]['corrected'][selection],
                                                        norm_pams['spectra_poly_degree'])
                    transmission[obs]['continuum'] = np.polynomial.chebyshev.chebval(
                        transmission['wave'], transmission[obs]['continuum_coeff'])

                    transmission[obs]['normalized'] = transmission[obs]['corrected'] / transmission[obs]['continuum']
                    transmission[obs]['normalized_err'] = transmission[obs]['corrected_err'] / \
                        transmission[obs]['continuum']

                    #import matplotlib.pyplot as plt
                    #plt.scatter(transmission['wave'], transmission[obs]['corrected'])
                    #plt.plot(transmission['wave'], transmission[obs]['continuum'])
                    #plt.scatter(transmission['wave'][selection], transmission[obs]['corrected'][selection], c='r')
                    #plt.scatter(transmission['wave'], transmission[obs]['corrected_err']+0.05, c='b')
                    #plt.scatter(transmission['wave'], transmission[obs]['normalized_err'], c='r')
                    #plt.show()
                    #quit()

                    transmission[obs]['continuum_uncorrected_coeff'] = \
                        np.polynomial.chebyshev.chebfit(transmission['wave'][selection],
                                                        transmission[obs]['rebinned'][selection],
                                                        norm_pams['spectra_poly_degree'])
                    transmission[obs]['continuum_uncorrected'] = np.polynomial.chebyshev.chebval(
                        transmission['wave'], transmission[obs]['continuum_uncorrected_coeff'])
                    transmission[obs]['normalized_uncorrected'] = transmission[obs]['rebinned'] / \
                        transmission[obs]['continuum_uncorrected']
                    transmission[obs]['normalized_uncorrected_err'] = transmission[obs]['rebinned_err'] / \
                        transmission[obs]['continuum_uncorrected']

                elif norm_pams['normalize_transmission']  and (
                    norm_pams['normalization_model'] == 'savgol'
                    or norm_pams['normalization_model'] == 'savitzky-golay'):

                    print(' ', obs, ' normalization using Savitzky-Golay filter')

                    transmission[obs]['continuum_coeff'] = None
                    transmission[obs]['continuum_uncorrected_coeff'] = None


                    transmission[obs]['continuum'] = savgol_filter(transmission[obs]['corrected'],
                                                            window_length=norm_pams['window_length'],
                                                            polyorder=norm_pams['polyorder'],
                                                            mode=norm_pams['mode'],
                                                            cval=norm_pams['cval'])

                    transmission[obs]['normalized'] = transmission[obs]['corrected'] / transmission[obs]['continuum']
                    transmission[obs]['normalized_err'] = transmission[obs]['corrected_err'] / \
                        transmission[obs]['continuum']

                    transmission[obs]['continuum_uncorrected'] = savgol_filter(transmission[obs]['rebinned'],
                                                            window_length=norm_pams['window_length'],
                                                            polyorder=norm_pams['polyorder'],
                                                            mode=norm_pams['mode'],
                                                            cval=norm_pams['cval'])

                    transmission[obs]['normalized_uncorrected'] = transmission[obs]['rebinned'] / transmission[obs]['continuum_uncorrected']
                    transmission[obs]['normalized_uncorrected_err'] = transmission[obs]['rebinned_err'] / \
                        transmission[obs]['continuum_uncorrected']

                else:
                    transmission[obs]['continuum_coeff'] = None
                    transmission[obs]['continuum'] = np.ones_like(transmission['wave'])
                    transmission[obs]['normalized'] = transmission[obs]['corrected'].copy()
                    transmission[obs]['normalized_err'] = transmission[obs]['corrected_err'].copy()

                    #import matplotlib.pyplot as plt
                    #plt.scatter(transmission['wave'], transmission[obs]['corrected'])
                    #plt.plot(transmission['wave'], transmission[obs]['continuum'])
                    #plt.scatter(transmission['wave'][selection], transmission[obs]['corrected'][selection], c='r')
                    # plt.show()

                    transmission[obs]['continuum_uncorrected_coeff'] = None
                    transmission[obs]['continuum_uncorrected'] = np.ones_like(transmission['wave'])
                    transmission[obs]['normalized_uncorrected'] = transmission[obs]['rebinned'].copy()
                    transmission[obs]['normalized_uncorrected_err'] = transmission[obs]['rebinned_err'].copy()
                    print_warning = False

            transm_average = np.zeros([len(lists['transit_full']), transmission['size']])
            weights_average = np.zeros([len(lists['transit_full']), transmission['size']])
            clvrm_average = np.zeros([len(lists['transit_full']), transmission['size']])
            uncorr_average = np.zeros([len(lists['transit_full']), transmission['size']])

            for i, obs in enumerate(lists['transit_full']):
                transm_average[i, :] = transmission[obs]['normalized'][:]
                weights_average[i, :] = 1./(transmission[obs]['normalized_err']**2.)
                clvrm_average[i, :] = transmission[obs]['clv_model_rebinned'][:]
                uncorr_average[i, :] = transmission[obs]['normalized_uncorrected'][:]

            transmission['average'], transmission['sum_weights'] = np.average(
                transm_average, axis=0, weights=weights_average, returned=True)
            transmission['average_err'] = 1. / np.sqrt(transmission['sum_weights'])

            transmission['average_clv_model'], _ = np.average(
                clvrm_average, axis=0, weights=weights_average, returned=True)

            transmission['average_uncorrected'], _ = np.average(
                uncorr_average, axis=0, weights=weights_average, returned=True)

            transmission['binned'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False)
            transmission['binned_err'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average_err'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False,
                               is_error=True)

            transmission['binned_clv_model'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average_clv_model'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False)

            transmission['binned_uncorrected'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average_uncorrected'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False)

            transm_average = np.zeros([len(lists['transit_out']), transmission['size']])
            weights_average = np.zeros([len(lists['transit_out']), transmission['size']])

            for i, obs in enumerate(lists['transit_out']):
                transm_average[i, :] = transmission[obs]['normalized'][:]
                weights_average[i, :] = 1./(transmission[obs]['normalized_err']**2.)

            transmission['average_out'], transmission['sum_weights_out'] = np.average(
                transm_average, axis=0, weights=weights_average, returned=True)
            transmission['average_out_err'] = 1./np.sqrt(transmission['sum_weights_out'])

            transmission['binned_out'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average_out'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False)
            transmission['binned_out_err'] = \
                rebin_1d_to_1d(transmission['wave'],
                               transmission['step'],
                               transmission['average_out_err'],
                               transmission['binned_wave'],
                               transmission['binned_step'],
                               preserve_flux=False,
                               is_error=True)

            #save_to_cpickle('transmission_'+reference+'_processed', processed, config_in['output'], night)
            save_to_cpickle(subroutine_name + '_' + reference + '_' + results_selection,
                            transmission, config_in['output'], night, lines_label, it_string)

            # Forcing memory deallocation
            transmission = None

        # Forcing memory deallocation
        clv_rm_models = None


def plot_transmission_spectrum(config_in, lines_label, night_input='', results_input='', reference='planetRF', pca_iteration=-1):

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]

    night_dict = from_config_get_nights(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    if results_input == '':
        results_list = ['user',
                        'mcmc_night_MED',
                        'mcmc_night_MAP',
                        'mcmc_global_MED',
                        'mcmc_global_MAP']
    else:
        results_list = np.atleast_1d(results_input)

    clv_rm_correction = lines_dict.get('clv_rm_correction', True)

    os.system('mkdir -p plots')

    interactive_plots = from_config_get_interactive_plots(config_in)

    for night in night_list:

        # Workaround to check if the transmission spectrum has been obtained through PCA iterations
        preparation_input = load_from_cpickle('transmission_preparation', config_in['output'], night)

        if preparation_input.get('pca_output', False):
            if pca_iteration >= 0:
                it_string = str(pca_iteration).zfill(2)
            else:
                it_string = str(preparation_input.get('ref_iteration', 0)).zfill(2)
        else:
            it_string = ''
        preparation_input = None

        if clv_rm_correction:
            try:
                clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night, lines_label)
            except (FileNotFoundError, IOError):
                clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night)

        for results_selection in results_list:

            filename_rad = subroutine_name + '_'+reference+'_'+results_selection

            """ Retrieving the list of observations"""
            lists = load_from_cpickle('lists', config_in['output'], night)

            """ Retrieving the analysis"""
            try:
                #processed = load_from_cpickle('transmission_'+reference+'_processed', config_in['output'], night)
                transmission = load_from_cpickle(filename_rad, config_in['output'], night, lines_label, it_string)
            except (FileNotFoundError, IOError):
                print()
                print("No transmission spectrum in {0:s}, no plots".format(reference))
                continue

            """ Creation of the color array, based on the BJD of the observations
            """
            bjd = []
            am = []

            for obs in lists['observations']:
                bjd.append(transmission[obs]['BJD'] - 2450000.0)
                am.append(transmission[obs]['AIRMASS'])

            color_cmap = plt.cm.viridis
            color_norm = plt.Normalize(vmin=bjd[0], vmax=bjd[-1])
            colors = color_cmap(color_norm(np.asarray(bjd)))

            fig = plt.figure(figsize=(12, 6))

            gs = GridSpec(2, 2, width_ratios=[50, 1])
            ax1 = plt.subplot(gs[0, 0])
            ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
            cbax1 = plt.subplot(gs[:, 1])

            # commented out because the plot was too cumbersome
            for obs in lists['transit_full']:

                color = [color_cmap(color_norm(transmission[obs]['BJD'] - 2450000.0))[:-1]]

                ax1.scatter(transmission['wave'],
                            transmission[obs]['normalized'],
                            c=color, s=1, zorder=3, alpha=0.25)

            for obs in lists['transit_out']:

                color = [color_cmap(color_norm(transmission[obs]['BJD'] - 2450000.0))[:-1]]

                ax2.scatter(transmission['wave'],
                            transmission[obs]['normalized'],
                            c=color, s=1, zorder=3, alpha=0.25)

            ax1.set_ylim(0.925, 1.075)
            ax2.set_xlabel('$\lambda$ [$\AA$]')
            ax2.legend(loc=3)
            ax1.set_title('Lines: {0:s} Night: {1:s} \n In-transit transmission spectrum in {2:s} \n Solution {3:s}'.format(
                lines_label, night, reference, results_selection))
            ax2.set_title('Out-transit transmission spectrum in {0:s}'.format(reference))
            try:
                ax1.set_xlim(lines_dict['plot_range'][0], lines_dict['plot_range'][1])
            except:
                ax1.set_xlim(lines_dict['range'][0], lines_dict['range'][1])

            sm = plt.cm.ScalarMappable(cmap=color_cmap, norm=color_norm)
            sm.set_array([])  # You have to set a dummy-array for this to work...
            cbar = plt.colorbar(sm, cax=cbax1)
            cbar.set_label('BJD - 2450000.0')
            fig.subplots_adjust(wspace=0.05, hspace=0.4)

            output_file = get_filename(filename_rad + '_observations',
                                       config_in['output'], night, lines_label, it_string, extension='.pdf')
            plt.savefig('plots/'+output_file, bbox_inches='tight', dpi=300)
            if interactive_plots:
                plt.show()
            plt.close()

            fig = plt.figure(figsize=(12, 6))

            gs = GridSpec(2, 2, width_ratios=[50, 1])
            ax1 = plt.subplot(gs[0, 0])
            ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
            cbax1 = plt.subplot(gs[:, 1])

            try:
                master_out = load_from_cpickle('master_out', config_in['output'], night)
                ax2.plot(master_out['wave'],
                         master_out['rescaled']-0.06,
                         color='k', zorder=10, label='master-out')
            except (FileNotFoundError, IOError):
                pass

            try:
                telluric = load_from_cpickle('telluric', config_in['output'], night)
                ax2.plot(telluric['template']['input']['wave'],
                         telluric['template']['input']['flux'] - 0.06,
                         color='C1', zorder=10, label='telluric')
                ax2.plot(telluric['template']['input']['wave'],
                         (telluric['template']['input']['flux']-1.)*10. + 1. - 0.06,
                         color='C2', alpha=0.5, zorder=9, label='telluric (x10)')
            except (FileNotFoundError, IOError, KeyError):
                pass

            #master_out = load_from_cpickle('master_out', config_in['output'], night)
            # ax1.errorbar(master_out['wave'],
            #            master_out['rescaled'],
            #            yerr=master_out['rescaled_err'],
            #            fmt='.', c='C0', label='master-out ' + night)

            ax1.errorbar(transmission['wave'],
                         transmission['average'],
                         yerr=transmission['average_err'],
                         fmt='ko', ms=1, zorder=5, alpha=0.25)

            ax1.errorbar(transmission['binned_wave'],
                         transmission['binned'],
                         yerr=transmission['binned_err'],
                         fmt='ro', ms=4, lw=2,  zorder=10)

            ax2.errorbar(transmission['wave'],
                         transmission['average_out'],
                         yerr=transmission['average_out_err'],
                         fmt='ko', ms=1, zorder=5, alpha=0.25, label='average')

            ax2.errorbar(transmission['binned_wave'],
                         transmission['binned_out'],
                         yerr=transmission['binned_out_err'],
                         fmt='ro', ms=4, lw=2,  zorder=10, label='binned average')

            ax1.set_ylim(0.99, 1.01)
            ax2.set_xlabel('$\lambda$ [$\AA$]')
            ax2.legend(loc=3)
            ax1.set_title('Lines: {0:s} Night: {1:s} \n In-transit transmission spectrum in {2:s} \n Solution {3:s}'.format(
                lines_label, night, reference, results_selection))
            ax2.set_title('Out-transit transmission spectrum in {0:s}'.format(reference))

            sm = plt.cm.ScalarMappable(cmap=color_cmap, norm=color_norm)
            sm.set_array([])  # You have to set a dummy-array for this to work...
            cbar = plt.colorbar(sm, cax=cbax1)
            cbar.set_label('BJD - 2450000.0')
            fig.subplots_adjust(wspace=0.05, hspace=0.4)

            try:
                ax1.set_xlim(lines_dict['plot_range'][0], lines_dict['plot_range'][1])
            except:
                ax1.set_xlim(lines_dict['range'][0], lines_dict['range'][1])

            #ax1.set_xlim(config_in['master-out']['wavelength_range'][0], config_in['master-out']['wavelength_range'][1])
            output_file = get_filename(filename_rad + '_binned',
                                       config_in['output'], night, lines_label, it_string, extension='.pdf')
            plt.savefig('plots/'+output_file, bbox_inches='tight', dpi=300)
            if interactive_plots:
                plt.show()
            plt.close()

            if not clv_rm_correction:
                continue

            fig = plt.figure(figsize=(12, 6))

            gs = GridSpec(2, 2, width_ratios=[50, 1])
            ax1 = plt.subplot(gs[0, 0])
            ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
            cbax1 = plt.subplot(gs[:, 1])

            # commented out because the plot was too cumbersome
            for obs in lists['transit_full']:

                color = [color_cmap(color_norm(transmission[obs]['BJD'] - 2450000.0))[:-1]]

                ax1.plot(clv_rm_models['common']['wave'],
                         transmission[obs]['clv_model_stellarRF'],
                         zorder=3, alpha=0.25)

                ax1.scatter(transmission['wave'],
                            transmission[obs]['clv_model_rebinned'],
                            c=color, s=1, zorder=10, alpha=0.5)

            for obs in lists['transit_out']:

                color = [color_cmap(color_norm(transmission[obs]['BJD'] - 2450000.0))[:-1]]

                ax2.plot(clv_rm_models['common']['wave'],
                         transmission[obs]['clv_model_stellarRF'],
                         zorder=3, alpha=0.25)

                ax2.scatter(transmission['wave'],
                            transmission[obs]['clv_model_rebinned'],
                            c=color, s=1, zorder=10, alpha=0.5)

            ax2.set_xlabel('$\lambda$ [$\AA$]')
            ax2.legend(loc=3)
            ax1.set_title('Lines: {0:s} Night: {1:s} \n CLV-RM correction in {2:s} \n Solution {3:s}'.format(
                lines_label, night, reference, results_selection))
            ax2.set_title('Out-transit transmission spectrum in {0:s}'.format(reference))
            try:
                ax1.set_xlim(lines_dict['plot_range'][0], lines_dict['plot_range'][1])
            except:
                ax1.set_xlim(lines_dict['range'][0], lines_dict['range'][1])

            sm = plt.cm.ScalarMappable(cmap=color_cmap, norm=color_norm)
            sm.set_array([])  # You have to set a dummy-array for this to work...
            cbar = plt.colorbar(sm, cax=cbax1)
            cbar.set_label('BJD - 2450000.0')
            fig.subplots_adjust(wspace=0.05, hspace=0.4)

            output_file = get_filename(filename_rad + '_clv_rm_models',
                                       config_in['output'], night, lines_label, it_string, extension='.pdf')
            plt.savefig('plots/'+output_file, bbox_inches='tight', dpi=300)
            if interactive_plots:
                plt.show()
            plt.close()



import numpy as np
from scipy.optimize import fsolve
import SLOPpy.subroutines.constants as constants

# +
# NAME:
#    exofast_keplereq
# PURPOSE:
#    Solve Kepler's Equation
# DESCRIPTION:
#    Solve Kepler's Equation. Method by S. Mikkola (1987) Celestial
#       Mechanics, 40 , 329-334.
#    result from Mikkola then used as starting value for
#       Newton-Raphson iteration to extend the applicability of this
#       function to higher eccentricities

__all__ = ["kepler_K1",
           "kepler_RV",
           "kepler_Tc2phase_Tref",
           "kepler_phase2Tc_Tref",
           "get_planet_mass",
           "kepler_true_anomaly_orbital_distance"]


def kepler_E(M_in, ec):
    E = 0.0
    E0 = 0.0
    M = np.atleast_1d(M_in)
    ecc = np.asarray(ec, dtype=np.double)
    eccanom = np.zeros(np.size(M), dtype=np.double)

    for ii in range(0, np.size(M)):
        # -np.pi < M < np.pi
        mx = M[ii]
        if mx > np.pi:
            mx = mx % (2. * np.pi)
            if mx > np.pi:
                mx = mx - (2. * np.pi)
        if mx <= -np.pi:
            mx = mx % (2. * np.pi)
            if mx < -np.pi:
                mx += (2. * np.pi)

        if ecc < 1e-10:
            eccanom[ii] = mx
        else:

            # equation 9a
            aux = 4.0 * ecc + 0.50
            alpha = (1.0 - ecc) / aux
            beta = mx / (2.0 * aux)

            # equation 9b
            ## the actual equation 9b is much much slower, but gives the same
            ## answer (probably because more refinement necessary)
            aux = np.sqrt(beta * beta + alpha * alpha * alpha)
            z = beta + aux
            if z < 0.:
                z = beta - aux
            z = z ** (1. / 3.)

            if abs(z) < 1e-8:
                s0 = 0.
            else:
                s0 = z - alpha / z

            s1 = s0 - (0.078 * s0 ** 5) / ((1.) + ecc)
            e0 = mx + ecc * (3. * s1 - 4. * s1 ** 3.)

            se0 = np.sin(e0)
            ce0 = np.cos(e0)

            f = e0 - ecc * se0 - mx
            f1 = (1.0) - ecc * ce0
            f2 = ecc * se0
            f3 = ecc * ce0
            u1 = -f / f1
            u2 = -f / (f1 + 0.5 * f2 * u1)
            u3 = -f / (f1 + 0.5 * f2 * u2 + (1. / 6.) * f3 * u2 ** 2.)
            u4 = -f / (f1 + 0.5 * f2 * u3 + (1. / 6.) * f3 * u3 ** 2 - (1. / 24.) * f2 * u3 ** 3)

            ecan_tmp = e0 + u4

            if ecan_tmp >= 2. * np.pi:
                ecan_tmp = ecan_tmp - 2. * np.pi

            if ecan_tmp < 0.:
                ecan_tmp = ecan_tmp + 2. * np.pi

            ## Now get more precise solution using Newton Raphson method
            ## for those times when the Kepler equation is not yet solved
            ## to better than 1e-10
            ## (modification J. Wilms)

            if mx < 0.:
                mx = mx + 2. * np.pi

            ## calculate the differences
            diff = abs(ecan_tmp - ecc * np.sin(ecan_tmp) - mx)
            if diff > abs(diff - 2 * np.pi):
                diff = abs(diff - 2 * np.pi)

            thresh1 = 1e-8
            thresh2 = 10000
            countt = 0

            while (diff > thresh1 and countt < thresh2):
                ## E-e sinE-M
                fe = (ecan_tmp - ecc * np.sin(ecan_tmp) - mx) % (2 * np.pi)
                ## f' = 1-e*cosE
                fs = (1. - ecc * np.cos(ecan_tmp)) % (2 * np.pi)
                oldval = ecan_tmp
                ecan_tmp = (oldval - fe / fs)
                diff = abs(oldval - ecan_tmp)
                countt += 1
            ## range reduction
            if ecan_tmp >= 2. * np.pi:
                ecan_tmp = ecan_tmp % 2. * np.pi
            if ecan_tmp < 0.:
                ecan_tmp = ecan_tmp % 2. * np.pi + 2. * np.pi

            eccanom[ii] = ecan_tmp

    return eccanom


def kepler_K1(m_star1, m_star2, period, i, e0):
    """ Computes the radial velocity semi-amplitude of the primary star

    :param m_star1: mass of the primary, in Solar mass units
    :param m_star2: mass of the secondary/planet, in Solar mass units
    :param period: orbital period of star2, in [d]
    :param i: orbital inclination of star2 wrt the observer (0=face on), in [deg]
    :param e0: orbital eccentricity of star2
    :return: k1, the observed radial velocity semi-amplitude of the primary, in [m s^-1]
    """
    # period must be given in days, conversion factor to seconds are included in the routine
    # constants.Gsi: Gravitational constant in SI system [m^3 kg^-1 s^-2]
    # constants.Msun: Sun mass in SI system [kg]
    # 86400. / constants.d2s: seconds in a day

    return (2. * np.pi * constants.Gsi * constants.Msun / 86400.) ** (1. / 3.) \
           * (np.sin(i * np.pi / 180.0) / np.sqrt(1. - e0 ** 2.)) * period ** (-1. / 3.) \
           * (m_star2 * (m_star1 + m_star2) ** (-2. / 3.))


def kepler_RV(BJD, TPeri, Period, gamma, K, e0, omega0):
    # omega = argument of pericenter
    # Mean Anomaly
    #
    MeAn = 2. * np.pi * (1. + (BJD - TPeri) / Period % 1.)

    if abs(e0) < 1e-3:
        TrAn = np.asarray(MeAn, dtype=np.double)
        e = np.asarray(0., dtype=np.double)
        omega = np.asarray(0., dtype=np.double)
    else:
        if e0 < 0.:
            e = np.asarray(-e0, dtype=np.double)
            omega = np.asarray(omega0, dtype=np.double) + np.pi
        else:
            e = np.asarray(e0, dtype=np.double)
            omega = np.asarray(omega0, dtype=np.double)

        # Eccentric Anomaly
        EccAn = kepler_E(MeAn, e)
        TrAn = 2. * np.arctan(np.sqrt((1.0 + e) / (1.0 - e)) * np.tan(EccAn / 2.0))

    rv = K * (np.cos(TrAn + omega) + e * np.cos(omega)) + gamma

    return rv


def kepler_RV_T0P(BJD0, phase, Period, K, e0, omega0):
    # BJD0 is given as BJD-T0, where T0 is arbitrarily defined by the user
    # Tperi_ is substituted by _phase_, which is the phase of the orbit where
    #        BJD0+T0+phase*Period = Tperi
    # omega = argument of pericenter
    #

    omega = np.asarray(omega0, dtype=np.double)
    e = np.asarray(e0, dtype=np.double)
    MeAn = 2. * np.pi * (1. + ((BJD0 / Period) + (phase - omega0) / (2 * np.pi)) % 1.)

    if abs(e0) < 1e-3:
        TrAn = np.asarray(MeAn, dtype=np.double)
        e = np.asarray(0., dtype=np.double)
    else:
        if e0 < 0.:
            e = -1 * e
            omega += np.pi

        # Eccentric Anomaly
        EccAn = kepler_E(MeAn, e)
        TrAn = 2. * np.arctan(np.sqrt((1.0 + e) / (1.0 - e)) * np.tan(EccAn / 2.0))

    rv = K * (np.cos(TrAn + omega) + e * np.cos(omega))

    return rv

def kepler_true_anomaly_orbital_distance(BJD0, Tcent0, Period, e0, omega0, a_sm):
    # BJD0 is given as BJD-T0, where T0 is arbitrarily defined by the user
    # Tperi_ is substituted by _phase_, which is the phase of the orbit where
    #        BJD0+T0+phase*Period = Tperi
    # omega = argument of pericenter

    phase = kepler_Tc2phase_Tref(Period, Tcent0, e0, omega0)

    omega = np.asarray(omega0, dtype=np.double)
    e = np.asarray(e0, dtype=np.double)
    MeAn = 2. * np.pi * (1. + ((BJD0 / Period) + (phase - omega0) / (2 * np.pi)) % 1.)

    if abs(e0) < 1e-3:
        TrAn = np.asarray(MeAn, dtype=np.double)
        e = np.asarray(0., dtype=np.double)
        r_orb = a_sm
    else:
        if e0 < 0.:
            e = -1 * e
            omega += np.pi

        # Eccentric Anomaly
        EccAn = kepler_E(MeAn, e)
        TrAn = 2. * np.arctan(np.sqrt((1.0 + e) / (1.0 - e)) * np.tan(EccAn / 2.0))
        r_orb = a_sm * (1. - e ** 2) / (1. + e * np.cos(TrAn))
    return TrAn, r_orb



def kepler_phase2Tc_Tref(Period, phase, e0, omega0):
    # The closest Tcent after Tref is given back
    TrAn = np.pi / 2 - omega0
    EccAn = 2. * np.arctan(np.sqrt((1.0 - e0) / (1.0 + e0)) * np.tan(TrAn / 2.0))
    MeAn = EccAn - e0 * np.sin(EccAn)
    return (MeAn - phase + omega0) / (2 * np.pi) * Period % Period


def kepler_Tc2phase_Tref(Period, Tcent, e0, omega0):
    # The closest Tcent after Tref is given back
    TrAn = np.pi / 2 - omega0
    EccAn = 2. * np.arctan(np.sqrt((1.0 - e0) / (1.0 + e0)) * np.tan(TrAn / 2.0))
    MeAn = EccAn - e0 * np.sin(EccAn)
    return (omega0 + MeAn - Tcent / Period * 2 * np.pi) % (2 * np.pi)


def f_get_mass(m_star2, m_star1, period, e0, k1):
    """ Computes the difference between the input radial velocity semi-amplitude
    of the primary star and the value corresponding to the provided orbital parameters.
    Supporting function to get_planet_mass subroutine

    :param m_star2: mass of the secondary/planet, in Solar mass units
    :param m_star1: mass of the primary, in Solar mass units
    :param period: orbital period of star2, in [d]
    :param e0: orbital eccentricity of star2
    :param k1: observed RV semi-amplitude of the primary
    :return: the difference between the observed and theoretical RV semi-amplitude of the primary, in [m s^-1]
    """
    # period must be given in days, conversion factor to seconds are included in the routine
    # constants.Gsi: Gravitational constant in SI system [m^3 kg^-1 s^-2]
    # constants.Msun: Sun mass in SI system [kg]
    # 86400. / constants.d2s: seconds in a day

    # M_star1, M_star2 in solar masses
    # P in days -> Period is converted in seconds in the routine
    # inclination assumed to be 90 degrees
    # Gravitational constant in SI system [in m^3 kg^-1 s^-2]
    # output in m/s

    return k1 \
           - ((2. * np.pi * constants.Gsi * constants.Msun / 86400.0) ** (1. / 3.)
              * (1. / np.sqrt(1. - e0 ** 2.))
              * period ** (-1. / 3.)
              * (m_star2 * (m_star1 + m_star2) ** (-2. / 3.)))

def get_approximate_mass(period, k1, e0, m_star1):
    """ Return the approximate mass of the planet in Solar mass units, in the assumption that M_planet << M_star

    :param period: orbital period of star2, in [d]
    :param k1: observed RV semi-amplitude of the primary
    :param e0: orbital eccentricity of star2
    :param m_star1: mass of the primary, in Solar mass units
    :return: mass of the planet, in Solar mass units
    """
    return k1 / ((2. * np.pi * constants.Gsi * constants.Msun / 86400.0) ** (1. / 3.)
         * (1. / np.sqrt(1. - e0 ** 2.))
         * period ** (-1. / 3.)
         * (m_star1 ** (-2. / 3.)))


def get_planet_mass(P, K, e, Mstar, approximation_limit=30.):

    n = np.size(K)
    if n == 1:
        M_approx = min(get_approximate_mass(P, K, e, Mstar), 2*constants.Msear)
        return fsolve(f_get_mass, M_approx, args=(Mstar, P, e, K))

    M_approx = get_approximate_mass(P, K, e, Mstar)

    if np.average(M_approx) > approximation_limit/constants.Msear:
        print('Computing exact mass of the planet (average approximate mass larger than {0:3.1f} Me)'.format(approximation_limit))
        M_init = np.average(M_approx)
        for i in range(0, n):
            M_approx[i] = fsolve(f_get_mass, np.average(M_init), args=(Mstar[i], P[i], e[i], K[i]))
    else:
        print('Computing planetary mass under the approximation M_planet << M_star (threshold at {0:3.1f} Me)'.format(approximation_limit))

    return M_approx



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.constants import *
from SLOPpy.subroutines.kepler_exo import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.math_functions import *

from astropy.convolution import Gaussian1DKernel, convolve

__all__ = ['compute_clv_rm_models_lines', 'plot_clv_rm_models_lines']

subroutine_name = 'clv_rm_models_lines'


def compute_clv_rm_models_lines(config_in, lines_label):
    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)

    planet_dict = from_config_get_planet(config_in)
    star_dict = from_config_get_star(config_in)
    clv_rm_dict = from_config_get_clv_rm(config_in)

    spectral_lines = from_config_get_spectral_lines(config_in)
    line_iter_dict = spectral_lines[lines_label]

    clv_rm_correction = line_iter_dict.get('clv_rm_correction', True)
    if not clv_rm_correction:
        return

    # wave_extension: additional range in wavelength added to avoid convolution
    # problems at the side of the spectrum
    wave_extension = 5.0
    # un-convolved portion of the spectrum given by range_boundaries +-
    # (wave_extension - wave_fix_convo)
    wave_fix_convo = 1.0

    # Added back-compatibility to old or "wrong" keys

    norm_dict = line_iter_dict.get('normalization', clv_rm_dict.get('normalization', {}))
    norm_pams={}
    norm_pams['model_poly_degree'] = norm_dict.get('model_poly_degree', 2)
    norm_pams['spectra_poly_degree'] = norm_dict.get('spectra_poly_degree', 2)
    norm_pams['lower_threshold'] = norm_dict.get('lower_threshold', 0.950)
    norm_pams['percentile_selection'] = norm_dict.get('percentile_selection', 10)

    try:
        synthesis = load_from_cpickle('clv_rm_synthesis', config_in['output'])
        star_grid = load_from_cpickle('clv_rm_star_grid', config_in['output'])

        #if not config_in['settings'].get('full_output', False):
        #    for night in night_dict:
        #        clv_rm_models = load_from_cpickle(
        #            'clv_rm_models', config_in['output'], night)

        print("{0:45s}                         {1:s}".format(
            subroutine_name, 'Retrieved'))


    except (FileNotFoundError, IOError):
        print("{0:45s}                         {1:s}".format(
            subroutine_name, 'Computing'))
        print()

        """
        Loading the spectral synthesis results, at the moment only SME output is supported. 
        Properties of the synthesis data files
        - limb_angles: this is an input to SME, so it is specific on how the synthesis has been performed
        - spectra: stellar spectrum as a function of the limb angle, sampled near the spectral lines
        - model: integrated spectrum of the star
        """

        synthesis_data_limb_angles = np.genfromtxt(
            clv_rm_dict['synthesis_files'] + '_muvals.txt', dtype=np.double)
        synthesis_data_spectra = np.genfromtxt(
            clv_rm_dict['synthesis_files'] + '_spectra.txt', dtype=np.double)
        synthesis_data_model = np.genfromtxt(
            clv_rm_dict['synthesis_files'] + '_model.txt', dtype=np.double)

        synthesis = {
            'surface': {
                'wave': synthesis_data_spectra[:, 0],
                'flux': synthesis_data_spectra[:, 1:],
                'n_mu': np.size(synthesis_data_limb_angles),
                'mu': synthesis_data_limb_angles
            },
            'total': {
                'wave': synthesis_data_model[:, 0],
                'norm': synthesis_data_model[:, 1],
            }
        }

        """ Setting up the array for model computation """

        synthesis['total']['step'] = synthesis['total']['wave'] * 0.0
        synthesis['total']['step'][1:] = synthesis['total']['wave'][1:] - \
            synthesis['total']['wave'][:-1]
        synthesis['total']['step'][0] = synthesis['total']['step'][1]

        synthesis['surface']['step'] = synthesis['surface']['wave'] * 0.0
        synthesis['surface']['step'][1:] = synthesis['surface']['wave'][1:] - \
            synthesis['surface']['wave'][:-1]
        synthesis['surface']['step'][0] = synthesis['surface']['step'][1]

        synthesis['surface']['wave_out'] = np.arange(synthesis['surface']['wave'][0],
                                                     synthesis['surface']['wave'][-1],
                                                     clv_rm_dict['rebinning_step'])
        synthesis['surface']['size_out'] = np.size(
            synthesis['surface']['wave_out'], axis=0)
        synthesis['surface']['step_out'] = np.ones(
            synthesis['surface']['size_out']) * clv_rm_dict['rebinning_step']

        synthesis['total']['norm_out'] = rebin_1d_to_1d(synthesis['total']['wave'],
                                                        synthesis['total']['step'],
                                                        synthesis['total']['norm'],
                                                        synthesis['surface']['wave_out'],
                                                        synthesis['surface']['step_out'],
                                                        method='exact_flux',
                                                        preserve_flux=False)

        """  Check if the number of spectra corresponds to the number of limb angle values """
        if np.size(synthesis['surface']['flux'], axis=1) != synthesis['surface']['n_mu']:
            print('ERROR in loading the stellar spectra')

        """
        Setting up the grid of stellar spectra for the CLV and RM computation
        odd number of points to include the zero value
        """
        star_grid = {
            'n_grid': clv_rm_dict['n_gridpoints'],
            'half_grid': int((clv_rm_dict['n_gridpoints'] - 1) / 2)
        }

        """ Coordinates of the centers of each grid cell (add offset) """
        star_grid['xx'] = np.linspace(-1.0000000000000, 1.0000000000000,
                                      star_grid['n_grid'], dtype=np.double)
        star_grid['xc'], star_grid['yc'] = np.meshgrid(
            star_grid['xx'], star_grid['xx'], indexing='xy')
        # check the Note section of the wiki page of meshgrid
        # https://docs.scipy.org/doc/numpy/reference/generated/numpy.meshgrid.html

        """ Distance of each grid cell from the center of the stellar disk """
        star_grid['rc'] = np.sqrt(star_grid['xc'] ** 2 + star_grid['yc'] ** 2)
        # Must avoid negative numbers inside the square root
        star_grid['inside'] = star_grid['rc'] < 1.0000000000000
        # Must avoid negative numbers inside the square root
        star_grid['outside'] = star_grid['rc'] >= 1.00000000000000

        """ Determine the mu angle for each grid cell, as a function of radius. """
        star_grid['mu'] = np.zeros([star_grid['n_grid'], star_grid['n_grid']],
                                   dtype=np.double)  # initialization of the matrix with the mu values
        star_grid['mu'][star_grid['inside']] = np.sqrt(
            1. - star_grid['rc'][star_grid['inside']] ** 2)

        """  2.2 Determine the Doppler shift to apply to the spectrum of each grid cell, from Cegla+2015 """

        star_grid['x_ortho'] = star_grid['xc'] * np.cos(star_dict['lambda'][0] * deg2rad) \
            - star_grid['yc'] * np.sin(
            star_dict['lambda'][0] * deg2rad)  # orthogonal distances from the spin-axis
        star_grid['y_ortho'] = star_grid['xc'] * np.sin(star_dict['lambda'][0] * deg2rad) \
            + star_grid['yc'] * np.cos(star_dict['lambda'][0] * deg2rad)

        star_grid['r_ortho'] = np.sqrt(
            star_grid['x_ortho'] ** 2 + star_grid['y_ortho'] ** 2)
        star_grid['z_ortho'] = np.zeros([star_grid['n_grid'], star_grid['n_grid']],
                                        dtype=np.double)  # initialization of the matrix
        star_grid['z_ortho'][star_grid['inside']] = np.sqrt(
            1. - star_grid['r_ortho'][star_grid['inside']] ** 2)

        """ rotate the coordinate system around the x_ortho axis by an agle: """
        star_grid['beta'] = (np.pi / 2.) - \
            star_dict['inclination'][0] * deg2rad

        """ orthogonal distance from the stellar equator """
        star_grid['yp_ortho'] = star_grid['z_ortho'] * np.sin(star_grid['beta']) + star_grid['y_ortho'] * np.cos(
            star_grid['beta'])

        """ stellar rotational velocity for a given position """
        star_grid['v_star'] = star_grid['x_ortho'] * star_dict['vsini'][0] * (
            1. - star_dict['alpha'][0] * star_grid['yp_ortho'] ** 2)
        # Null velocity for points outside the stellar surface
        star_grid['v_star'][star_grid['outside']] = 0.0

        """ Associate a synthetic spectrum to each cell """

    """ recomputation of spectra_mu  - most likely it has been deleted from the
    output file
    """

    star_grid['spectra_mu'] = [[0] * star_grid['n_grid']
                                for i in range(star_grid['n_grid'])]

    for x in range(0, star_grid['n_grid']):
        for y in range(0, star_grid['n_grid']):

            if star_grid['outside'][y, x]:
                continue

            index_closer = np.abs(
                synthesis['surface']['mu'] - star_grid['mu'][y, x]).argmin()  # take the index of the closer value

            if star_grid['mu'][y, x] in synthesis['surface']['mu']:
                star_grid['spectra_mu'][x][y] = synthesis['surface']['flux'][:, index_closer]
                continue
            elif index_closer == synthesis['surface']['n_mu'] - 1 or \
                    synthesis['surface']['mu'][index_closer] > star_grid['mu'][y, x]:
                mu_ind0 = index_closer - 1
                mu_ind1 = index_closer
            else:
                mu_ind0 = index_closer
                mu_ind1 = index_closer + 1

            diff_mu = synthesis['surface']['mu'][mu_ind1] - \
                synthesis['surface']['mu'][mu_ind0]

            star_grid['spectra_mu'][x][y] = synthesis['surface']['flux'][:, mu_ind0] \
                + (star_grid['mu'][y, x] - synthesis['surface']['mu'][mu_ind0]) / diff_mu \
                * (synthesis['surface']['flux'][:, mu_ind1]
                   - synthesis['surface']['flux'][:, mu_ind0])

    """ Computation of the continuum level (total flux is already normalized)"""
    star_grid['continuum'] = [[0] * star_grid['n_grid']
                              for i in range(star_grid['n_grid'])]

    spectral_window = ((synthesis['surface']['wave'] > clv_rm_dict['continuum_range'][0]) &
                       (synthesis['surface']['wave'] < clv_rm_dict['continuum_range'][1]))

    for x in range(0, star_grid['n_grid']):
        for y in range(0, star_grid['n_grid']):

            if star_grid['outside'][y, x]:
                continue
            star_grid['continuum'][x][y] = np.median(
                star_grid['spectra_mu'][x][y][spectral_window])

    star_grid['continuum_level'] = np.sum(star_grid['continuum'])

    """
    Setting up the grid for the rescaling factor of the planetary radius
    """
    try:
        radius_grid = np.arange(clv_rm_dict['radius_factor'][0],
                                clv_rm_dict['radius_factor'][1] +
                                clv_rm_dict['radius_factor'][2],
                                clv_rm_dict['radius_factor'][2])
    except KeyError:
        radius_grid = np.arange(0.5, 2.6, 0.1)

    """ CLV + RM model computation is performed only on the wavelength range of
    interest, with the addition of a few Angstroms """

    wave_selection = (synthesis['surface']['wave_out'] > line_iter_dict['range'][0]-wave_extension) \
        & (synthesis['surface']['wave_out'] < line_iter_dict['range'][1]+wave_extension)

    for night in night_dict:
        """ Retrieving the list of observations"""

        print()
        print("compute_CLV_RM_models for lines {0:s},   Night: {1:s}".format(lines_label, night))

        try:
            clv_rm_models = load_from_cpickle(
                'clv_rm_models', config_in['output'], night, lines_label)
            continue
        except:
            print()
            print("  No CLV & RM correction files found for lines {0:s}, Night: {1:s} , computing now".format(lines_label, night))

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle(
            'observational_pams', config_in['output'], night)

        instrument = night_dict[night]['instrument']

        clv_rm_models = {
            'common': {
                'wave': synthesis['surface']['wave_out'][wave_selection],
                'step': synthesis['surface']['step_out'][wave_selection],
                'norm': synthesis['total']['norm_out'][wave_selection],
                'size': int(np.sum(wave_selection)),
                'continuum_level': star_grid['continuum_level'],
                'radius_grid': radius_grid,
                'n_radius_grid': len(radius_grid)
            }
        }

        clv_rm_models['common']['convolution_dlambda'] = \
            np.median(clv_rm_models['common']['wave']) / \
            instrument_dict[instrument]['resolution']
        clv_rm_models['common']['convolution_sigma'] = \
            clv_rm_models['common']['convolution_dlambda'] / \
            np.median(clv_rm_models['common']['step'])
        gaussian = Gaussian1DKernel(
            stddev=clv_rm_models['common']['convolution_sigma'])

        clv_rm_models['common']['norm_convolved'] = convolve(
            clv_rm_models['common']['norm'], gaussian)

        """ Fixing border effect (we took already wave_extension angstrom outside of the
        actual range, so doing it this way is fine)"""
        wave_fix = wave_extension - wave_fix_convo

        wave_fix_convolution = (clv_rm_models['common']['wave'] < line_iter_dict['range'][0]-wave_fix) | (clv_rm_models['common']['wave'] > line_iter_dict['range'][1]+wave_fix) 
        clv_rm_models['common']['norm_convolved'][wave_fix_convolution] = clv_rm_models['common']['norm'][wave_fix_convolution]

        #import matplotlib.pyplot as plt
        #plt.plot(clv_rm_models['common']['wave'], clv_rm_models['common']['norm'], c='C1')
        #plt.plot(clv_rm_models['common']['wave'], clv_rm_models['common']['norm_convolved'], c='C2')
        #plt.show()
        #quit()

        """
        Computation of the first derivative, useful to identify
        continuum level. This method is prone to errors for
        observational data, but it's quite robust for synthetic spectra
        if jumps in wavelngth are small
        """

        clv_rm_models['common']['norm_convolved_derivative'] = \
                first_derivative(clv_rm_models['common']['wave'],
                                    clv_rm_models['common']['norm_convolved'])

        wave_fix = wave_extension - wave_fix_convo*2
        exclude_borders = (clv_rm_models['common']['wave'] > line_iter_dict['range'][0]+wave_fix) & (clv_rm_models['common']['wave'] < line_iter_dict['range'][1]-wave_fix)

        print('  Range for  continuum normalization: ',line_iter_dict['range'][0]-wave_fix, line_iter_dict['range'][1]+wave_fix)

        # Using only the 10percentile of values of the derivative around zero
        cont_10perc = np.percentile(np.abs(clv_rm_models['common']['norm_convolved_derivative']), norm_pams['percentile_selection'])

        clv_rm_models['common']['norm_convolved_bool'] = (np.abs(clv_rm_models['common']['norm_convolved_derivative']) < cont_10perc) \
            & (exclude_borders) \
            & (clv_rm_models['common']['norm_convolved']> norm_pams['lower_threshold'])

        print('  Number of points within 10percentile: {0:10.0f}'.format(np.sum((np.abs(clv_rm_models['common']['norm_convolved_derivative']) < cont_10perc))))
        print('  Number of points included in restricted borders: {0:10.0f}'.format(np.sum(exclude_borders)))
        print('  Number of points above threshold: {0:10.0f}'.format(np.sum( (clv_rm_models['common']['norm_convolved']> norm_pams['lower_threshold']))))

        norm_convolved_bool = (np.abs(clv_rm_models['common']['norm_convolved_derivative']) < cont_10perc) \
            & (exclude_borders) \
            & (clv_rm_models['common']['norm_convolved']> norm_pams['lower_threshold'])

        if np.sum(norm_convolved_bool) < np.sum(exclude_borders)/50:
            print('  Lower threshold decreased by 80% to allow point selection ', norm_pams['lower_threshold']*0.80)

            clv_rm_models['common']['norm_convolved_bool'] = (np.abs(clv_rm_models['common']['norm_convolved_derivative']) < cont_10perc) \
                & (exclude_borders) & (clv_rm_models['common']['norm_convolved']> norm_pams['lower_threshold']*0.80)
        else:
            clv_rm_models['common']['norm_convolved_bool'] = norm_convolved_bool


        print('  Number of points for continuum normalization: {0:10.0f}'.format(np.sum(clv_rm_models['common']['norm_convolved_bool'])))

        processed = {}

        print()
        for obs in lists['observations']:

            print('  Computing CLV+RM correction for ', obs)

            processed[obs] = {}
            clv_rm_models[obs] = {}

            n_oversampling = int(
                observational_pams[obs]['EXPTIME'] / clv_rm_dict['time_step'])

            if n_oversampling % 2 == 0:
                n_oversampling += 1
            half_time = observational_pams[obs]['EXPTIME'] / 2 / 86400.

            processed[obs]['bjd_oversampling'] = np.linspace(observational_pams[obs]['BJD'] - half_time,
                                                             observational_pams[obs]['BJD'] + half_time,
                                                             n_oversampling, dtype=np.double)


            if planet_dict['orbit'] == 'circular':
                # Time of pericenter concides with transit time, if we assume e=0 and omega=np.pi/2.
                eccentricity = 0.00
                omega_rad = np.pi / 2.
                # Tcent is assumed as reference time
                Tref = planet_dict['reference_time_of_transit'][0]
                Tcent_Tref = 0.000
            else:
                omega_rad = planet_dict['omega'][0] * deg2rad
                Tref = planet_dict['reference_time']
                Tcent_Tref = planet_dict['reference_time_of_transit'][0] - Tref
                eccentricity = planet_dict['eccentricity'][0]
            inclination_rad = planet_dict['inclination'][0] * deg2rad

            true_anomaly, orbital_distance_ratio = kepler_true_anomaly_orbital_distance(
                processed[obs]['bjd_oversampling'] - Tref,
                Tcent_Tref,
                planet_dict['period'][0],
                eccentricity,
                omega_rad,
                planet_dict['semimajor_axis_ratio'][0])

            """ planet position during its orbital motion, in unit of stellar radius"""
            # Following Murray+Correia 2011 , with the argument of the ascending node set to zero.
            # 1) the ascending node coincide with the X axis
            # 2) the reference plance coincide with the plane of the sky
            processed[obs]['planet_position'] = {
                'xp': -orbital_distance_ratio * (np.cos(omega_rad + true_anomaly)),
                'yp': orbital_distance_ratio * (np.sin(omega_rad + true_anomaly) * np.cos(inclination_rad)),
                'zp': orbital_distance_ratio * (np.sin(inclination_rad) * np.sin(omega_rad + true_anomaly))
            }
            # projected distance of the planet's center to the stellar center
            processed[obs]['planet_position']['rp'] = np.sqrt(processed[obs]['planet_position']['xp'] ** 2
                                                              + processed[obs]['planet_position']['yp'] ** 2)

            # obscured flux integrated over the full epoch
            # grid n_radius_grid X size_out (of spectral model)
            clv_rm_models[obs]['missing_flux'] = np.zeros(
                [len(radius_grid), clv_rm_models['common']['size']], dtype=np.double)

            # iterating on the sub-exposures
            for j, zeta in enumerate(processed[obs]['planet_position']['zp']):

                if zeta > 0 and processed[obs]['planet_position']['rp'][j] < 1. + planet_dict['radius_ratio'][0]:
                    # the planet is in the foreground or inside the stellar disk, continue
                    # adjustment: computation is performed even if only part of the planet is shadowing the star

                    rd = np.sqrt((processed[obs]['planet_position']['xp'][j] - star_grid['xc']) ** 2 +
                                 (processed[obs]['planet_position']['yp'][j] - star_grid['yc']) ** 2)

                    # iterating on the cell grid
                    for x in range(0, star_grid['n_grid']):
                        for y in range(0, star_grid['n_grid']):

                            # skip the step if the cell is outside the stellar disk
                            # or if the cell is not shadowed by the planet when the largest possible size is considered
                            if star_grid['outside'][y, x] or rd[y, x] > planet_dict['radius_ratio'][0]*radius_grid[-1]:
                                continue



                            # rescaled planetary radius selection
                            grid_sel = (
                                rd[y, x] <= planet_dict['radius_ratio'][0]*radius_grid)
                            # stellar flux in the masked region
                            flux_tmp = rebin_1d_to_1d(synthesis['surface']['wave'],
                                                      synthesis['surface']['step'],
                                                      star_grid['spectra_mu'][x][y],
                                                      clv_rm_models['common']['wave'],
                                                      clv_rm_models['common']['step'],
                                                      rv_shift=star_grid['v_star'][y, x],
                                                      method='exact_flux',
                                                      preserve_flux=False)

                            # fixing zero values that may have been introduced by
                            # the rebinning process from an extremely irregular sampling

                            ind_sel = np.where(flux_tmp < 0.)[0]
                            for ii in ind_sel:
                                if ii == 0:
                                    flux_tmp[ii] = flux_tmp[ii + 1]
                                elif ii == np.size(flux_tmp) - 1:
                                    flux_tmp[ii] = flux_tmp[ii - 1]
                                else:
                                    flux_tmp[ii] = (
                                        flux_tmp[ii - 1] + flux_tmp[ii + 1]) / 2.

                            """
                            Outer product of the radius selection array (size=M)
                            and the flux array (N) so that it can be summed
                            properly to the MxN missing_flux matrix.
                            """
                            clv_rm_models[obs]['missing_flux'] += \
                                np.outer(grid_sel, flux_tmp)

            clv_rm_models[obs]['missing_flux'] /= n_oversampling
            clv_rm_models[obs]['stellar_spectra'] = \
                np.outer(np.ones(len(radius_grid)), clv_rm_models['common']['norm']) \
                - (clv_rm_models[obs]['missing_flux'] /
                   clv_rm_models['common']['continuum_level'])

            clv_rm_models[obs]['stellar_spectra_convolved'] = \
                np.zeros([len(radius_grid), clv_rm_models['common']['size']],
                         dtype=np.double)

            clv_rm_models[obs]['clv_rm_model_convolved'] = \
                np.zeros([len(radius_grid), clv_rm_models['common']['size']],
                         dtype=np.double)

            clv_rm_models[obs]['clv_rm_model_convolved_derivative'] = \
                np.zeros([len(radius_grid), clv_rm_models['common']['size']],
                         dtype=np.double)

            clv_rm_models[obs]['clv_rm_model_convolved_continuum_bool'] = \
                np.zeros([len(radius_grid), clv_rm_models['common']['size']],
                         dtype=bool)

            clv_rm_models[obs]['clv_rm_model_convolved_normalized'] = \
                np.zeros([len(radius_grid), clv_rm_models['common']['size']],
                         dtype=np.double)

            for ii in range(0, len(radius_grid)):
                clv_rm_models[obs]['stellar_spectra_convolved'][ii, :] = \
                    convolve(clv_rm_models[obs]['stellar_spectra'][ii, :],
                             gaussian)

                clv_rm_models[obs]['stellar_spectra_convolved'][ii, wave_fix_convolution] = \
                    clv_rm_models[obs]['stellar_spectra'][ii, wave_fix_convolution]


                """
                This is the theoretical transmission spectrum in the stellar reference frame
                when only CLV and RM effects are present (no atmospheric
                transmission)
                """
                clv_rm_models[obs]['clv_rm_model_convolved'][ii, :] = \
                    clv_rm_models[obs]['stellar_spectra_convolved'][ii, :] \
                    / clv_rm_models['common']['norm_convolved']

                """
                High-resolution transmission spectra are always rescaled for
                their continuum because in fiber-fed spectrographs the
                information on the absolute flux of the star is lost.

                If not using the normalized spectrum, normalization factor must
                be included somehow when correcting for the CLV+RM, before
                fitting the atomic absoprtion lines
                """

                normalization_function = np.polynomial.chebyshev.Chebyshev.fit(
                    clv_rm_models['common']['wave'][clv_rm_models['common']['norm_convolved_bool']],
                    clv_rm_models[obs]['clv_rm_model_convolved'][ii, :][clv_rm_models['common']['norm_convolved_bool']],
                    deg=norm_pams['model_poly_degree']
                )

                clv_rm_models[obs]['clv_rm_model_convolved_normalized'][ii, :] = clv_rm_models[obs]['clv_rm_model_convolved'][ii, :] / normalization_function(clv_rm_models['common']['wave'])

                ##if ii!= 5: continue
                ##import matplotlib.pyplot as plt
                ##plt.plot(clv_rm_models['common']['wave'], clv_rm_models[obs]['stellar_spectra_convolved'][ii, :], c='C0')
                ##plt.plot(clv_rm_models['common']['wave'], clv_rm_models['common']['norm_convolved'], c='C1')
                ##plt.show()
                ##
                ##
                ##print(np.sum(clv_rm_models['common']['norm_convolved_bool']))
                ##plt.plot(clv_rm_models['common']['wave'], clv_rm_models[obs]['clv_rm_model_convolved_normalized'][ii, :], label='convolved, normalized')
                ##plt.plot(clv_rm_models['common']['wave'], clv_rm_models[obs]['clv_rm_model_convolved'][ii, :], label='convolved')
                ##plt.plot(clv_rm_models['common']['wave'], normalization_function(clv_rm_models['common']['wave']), c='C5', zorder=10)
                ##plt.scatter(
                ##    clv_rm_models['common']['wave'][clv_rm_models['common']['norm_convolved_bool']],
                ##    clv_rm_models[obs]['clv_rm_model_convolved'][ii, :][clv_rm_models['common']['norm_convolved_bool']], s=5, c='C3')
                ##plt.legend()
                ##plt.show()
                ##if obs=='HARPN.2016-06-04T02-27-41.158': quit()


            """ In the planetary reference frame, the corrected transmission
            spectrum T_corr is given by

            T_corr = T_input * (synthetic_convolved /
            stellar_spectra_convolved),

            where: T_input: transmission spectrum before the correction
            synthetic_convolved: integrated synthetic stellar spectrum,
            convolved for the instrumental resolution.
            stellar_spectra_convolved: stellar spectrum after removing the
            contribute of the stellar surface covered by the planet, convolved
            for the instrumental resolution (synthetic_convolved and
            stellar_spectra_convolved are in the stellar rest frame must be
            rebinned in the planetary rest frame)

            Since clv_rm_model_convolved = stellar_spectra_convolved /
            synthetic_convolved the observed transmission spectrum must be
            DIVIDED by clv_rm_model_convolved
            """

        save_to_cpickle('clv_rm_models', clv_rm_models,
                        config_in['output'], night, lines_label)
        clv_rm_models = None  # Forcing memory de-allocation

    if not config_in['settings'].get('full_output', False):
        del star_grid['spectra_mu']


    try:
        synthesis = load_from_cpickle('clv_rm_synthesis', config_in['output'])
        star_grid = load_from_cpickle('clv_rm_star_grid', config_in['output'])
    except (FileNotFoundError, IOError):
        save_to_cpickle('clv_rm_star_grid', star_grid, config_in['output'])
        save_to_cpickle('clv_rm_synthesis', synthesis, config_in['output'])


def plot_clv_rm_models_lines(config_in, lines_label, night_input=''):

    spectral_lines = from_config_get_spectral_lines(config_in)
    line_iter_dict = spectral_lines[lines_label]

    clv_rm_correction = line_iter_dict.get('clv_rm_correction', True)
    if not clv_rm_correction:
        return

    night_dict = from_config_get_nights(config_in)

    synthesis = load_from_cpickle('clv_rm_synthesis', config_in['output'])
    star_grid = load_from_cpickle('clv_rm_star_grid', config_in['output'])


    if night_input == '':
        # Visualize the mu of star
        fig = plt.figure(figsize=(8, 6.5))

        plt.title('Limb angle')
        plt.contourf(star_grid['xx'], star_grid['xx'],
                     star_grid['mu'], 60, cmap=plt.cm.viridis)
        plt.colorbar(label='$\mu$')  # draw colorbar
        # plot data points.
        plt.xlim(-1, 1)
        plt.ylim(-1, 1)
        plt.xlabel('x [R_s]')
        plt.ylabel('y [R_s]')
        plt.show()

        # Visualize the RV of star
        fig = plt.figure(figsize=(8, 6.5))

        # CS = plt.contour(xx,xx,v_star,50,linewidths=0.5,colors='k')
        plt.title('Radial velocity field')

        plt.contourf(star_grid['xx'], star_grid['xx'],
                     star_grid['v_star'], 100, cmap=plt.cm.seismic)
        plt.colorbar(label='v_star')  # draw colorbar
        plt.xlim(-1, 1)
        plt.ylim(-1, 1)
        plt.xlabel('x [R_s]')
        plt.ylabel('y [R_s]')
        plt.show()

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        lists = load_from_cpickle('lists', config_in['output'], night)
        clv_rm_models = load_from_cpickle(
            'clv_rm_models', config_in['output'], night, lines_label)
        observational_pams = load_from_cpickle(
            'observational_pams', config_in['output'], night)

        colors_properties, colors_plot, colors_scatter = make_color_array_matplotlib3(
            lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        i0_radius = np.argmin(
            np.abs(clv_rm_models['common']['radius_grid']-1.00))

        for obs in lists['transit_in']:
            ax1.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['stellar_spectra'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

            ax1.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['missing_flux'][i0_radius, :] /
                     clv_rm_models['common']['continuum_level'],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

            ax2.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['stellar_spectra'][-1, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

            ax2.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['missing_flux'][-1, :] /
                     clv_rm_models['common']['continuum_level'],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

        # for obs in lists['transit_out']:
        #    ax2.plot(clv_rm_models['common']['wave'],
        #             clv_rm_models[obs]['stellar_spectra'],
        #             color=colors_plot['mBJD'][obs], alpha=0.2)

        ax1.set_title(
            'Night: {0:s} \n Input spectra, stellar radius'.format(night))
        ax2.set_title('Stellar radius x {0:2.2f}'.format(
            clv_rm_models['common']['radius_grid'][-1]))
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(
            cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()









        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        i0_radius = np.argmin(
            np.abs(clv_rm_models['common']['radius_grid']-1.00))

        for obs in lists['transit_out']:
            ax1.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['stellar_spectra'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

            ax2.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['stellar_spectra_convolved'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)


        ax1.plot(clv_rm_models['common']['wave'],
                    clv_rm_models['common']['norm'][:],
                    color='C3')

        ax2.plot(clv_rm_models['common']['wave'],
                    clv_rm_models['common']['norm_convolved'][:],
                    color='C3')

        # for obs in lists['transit_out']:
        #    ax2.plot(clv_rm_models['common']['wave'],
        #             clv_rm_models[obs]['stellar_spectra'],
        #             color=colors_plot['mBJD'][obs], alpha=0.2)

        ax1.set_title(
            'Night: {0:s} \n CLV+RM correction, convolved '.format(night))
        ax2.set_title('Out of transit')
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(
            cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()







        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        i0_radius = np.argmin(
            np.abs(clv_rm_models['common']['radius_grid']-1.00))

        for obs in lists['transit_in']:
            ax1.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['clv_rm_model_convolved'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

        for obs in lists['transit_out']:

            ax2.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['clv_rm_model_convolved'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

        # for obs in lists['transit_out']:
        #    ax2.plot(clv_rm_models['common']['wave'],
        #             clv_rm_models[obs]['stellar_spectra'],
        #             color=colors_plot['mBJD'][obs], alpha=0.2)

        ax1.set_title(
            'Night: {0:s} \n CLV+RM correction, convolved '.format(night))
        ax2.set_title('Out of transit')
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(
            cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()











        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        i0_radius = np.argmin(
            np.abs(clv_rm_models['common']['radius_grid']-1.00))

        for obs in lists['transit_in']:
            #ax1.plot(clv_rm_models['common']['wave'],
            #         clv_rm_models[obs]['clv_rm_model_convolved'][i0_radius, :],
            #         color=colors_plot['mBJD'][obs], alpha=0.2)

            ax1.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['clv_rm_model_convolved_normalized'][i0_radius, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)


        for obs in lists['transit_out']:

            #ax2.plot(clv_rm_models['common']['wave'],
            #         clv_rm_models[obs]['clv_rm_model_convolved'][-1, :],
            #         color=colors_plot['mBJD'][obs], alpha=0.2)

            ax2.plot(clv_rm_models['common']['wave'],
                     clv_rm_models[obs]['clv_rm_model_convolved_normalized'][-1, :],
                     color=colors_plot['mBJD'][obs], alpha=0.2)

        # for obs in lists['transit_out']:
        #    ax2.plot(clv_rm_models['common']['wave'],
        #             clv_rm_models[obs]['stellar_spectra'],
        #             color=colors_plot['mBJD'][obs], alpha=0.2)

        ax1.set_title(
            'Night: {0:s} \n CLV+RM correction, convolved and normalized '.format(night))
        ax2.set_title('Out of transit')
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(
            cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()



from __future__ import print_function, division
import numpy as np
try:
    import cPickle as pickle
except:
    import pickle

import os
from os import path
import oyaml as yaml
from SLOPpy.subroutines.object_parameters import StarParameters, PlanetParameters
from SLOPpy.config_default import *

__all__ = ["save_to_cpickle",
           "load_from_cpickle",
           "delete_cpickle",
           "check_existence_cpickle",
           "get_filename",
           "load_yaml_file",
           "pars_input",
           "yaml_parser",
           "get_filelists",
           "from_config_get_nights",
           "from_config_get_instrument",
           "from_config_get_system",
           "from_config_get_pipeline",
           "from_config_get_planet",
           "from_config_get_star",
           "from_config_get_clv_rm",
           "from_config_refraction",
           "from_config_get_interstellar_lines",
           "from_config_get_transmission_lightcurve",
           "from_config_get_transmission",
           "from_config_get_molecfit",
           "from_config_get_transmission_mcmc",
           "from_config_get_spectral_lines",
           "from_config_get_interactive_plots",
           "from_config_get_pca_parameters",
           "from_config_get_fullspectrum_parameters"]

accepted_extensions = ['.yaml', '.yml', '.conf', '.config', '.input', ]


def save_to_cpickle(fname, dictionary, output, night='', lines='', it_string='', temp_folder=''):

    output_file = get_filename(fname, output, night, lines, it_string, temp_folder=temp_folder)
    pickle.dump(dictionary, open(output_file, "wb"))


def load_from_cpickle(fname, output, night='', lines='', it_string='', temp_folder=''):

    output_file = get_filename(fname, output, night, lines, it_string, temp_folder=temp_folder)
    return pickle.load(open(output_file, "rb"))


def delete_cpickle(fname, output, night='', lines='', it_string='', temp_folder=''):

    output_file = get_filename(fname, output, night, lines, it_string, temp_folder=temp_folder)
    os.remove(output_file)

def check_existence_cpickle(fname, output, night='', lines='', it_string='', temp_folder=''):

    output_file = get_filename(fname, output, night, lines, it_string, temp_folder=temp_folder)
    return path.isfile(output_file)


def get_filename(fname, output, night, lines='', it_string='', extension=".p", temp_folder=''):

    if len(temp_folder) > 0:
        str_lines = temp_folder + '/'+ output
    else:
        str_lines = output

    for str_input in [lines, night, fname, it_string]:
        if len(str_input) > 0:
            str_lines += '_' + str_input
    return str_lines + extension

    if lines == '':
        if night == '':
            return output + '_' + fname + extension
        else:
            return output + '_' + night + "_" + fname + extension
    else:
        if night == '':
            return output + '_' + lines + '_' + fname + extension
        else:
            return output + '_' + lines + '_' + night + "_" + fname + extension


def load_yaml_file(file_conf):
    # shortcut for jupyter notebook plots
    config_in = yaml_parser(file_conf)
    return pars_input(config_in)

def yaml_parser(file_conf):
    stream = open(file_conf, 'r')

    try:
        config_in = yaml.load(stream, Loader=yaml.FullLoader)
    except AttributeError:
        config_in = yaml.load(stream)
        print(' Consider updating YAML')
    except:
        print(' Some error happened while reading the configuration file')
        quit()

    if 'output' not in config_in:

        for extension in accepted_extensions:
            if file_conf.find(extension) > 0:
                output_name = file_conf.replace(extension, "")
                continue

        config_in['output'] = output_name

    return config_in


def pars_input(config_in):
    config_in['system'] = {}

    if 'settings' not in config_in:
        config_in['settings'] = config_default['settings'].copy()
    else:
        for key, key_val in config_default['settings'].items():
            if key not in config_in['settings']:
                 config_in['settings'][key] = key_val


    for instrument in config_in['instruments']:

        for key, key_val in config_default['instruments'].items():
            if key not in config_in['instruments'][instrument]:
                 config_in['instruments'][instrument][key] = key_val

        """ create the refraction dictionary if not listed under the instrument section"""
        if 'refraction' not in config_in['instruments'][instrument]:
            config_in['instruments'][instrument]['refraction'] = {}

        """ when the refractions parameters are not explicitely specified in this section, they are either inherited
        from the top level dictionary or copied from the default dictionary """
        for key, key_val in config_default['refraction'].items():
            if key not in config_in['instruments'][instrument]['refraction']:
                try:
                    config_in['instruments'][instrument]['refraction'][key] = config_in['refraction'][key]
                except:
                    config_in['instruments'][instrument]['refraction'][key] = key_val

    if 'master-out' not in config_in:
        config_in['master-out'] = config_default['master-out'].copy()
    else:
        for key, key_val in config_default['master-out'].items():
            if key not in config_in['master-out']:
                 config_in['master-out'][key] = key_val

    if 'shared' not in config_in['instruments']:
        if 'wavelength_step' not in config_in['master-out']:
            config_in['instruments']['shared'] = {
                'wavelength_step': 0.0100
            }
        else:
            config_in['instruments']['shared'] = {
                'wavelength_step': config_in['master-out']['wavelength_step']
            }

    if 'molecfit' not in config_in:
        config_in['molecfit'] = config_default['molecfit'].copy()
    else:
        for key, key_val in config_default['molecfit'].items():
            if key not in config_in['molecfit']:
                 config_in['molecfit'][key] = key_val

    if 'molecfit' not in config_in:
        config_in['molecfit'] = config_default['molecfit'].copy()
    else:
        for key, key_val in config_default['molecfit'].items():
            if key not in config_in['molecfit']:
                 config_in['molecfit'][key] = key_val



    for night in config_in['nights']:

        instrument = config_in['nights'][night]['instrument']

        """ keywords are inherited from the instrument dictionary, when not explicitely specified"""
        for key in copy_from_instrument:
            if key not in config_in['nights'][night]:
                try:
                    config_in['nights'][night][key] = config_in['instruments'][instrument][key]
                except KeyError:
                    continue

        if 'refraction' not in config_in['nights'][night]:
            config_in['nights'][night]['refraction'] = config_in['instruments'][instrument]['refraction'].copy()
        else:
            for key, key_val in config_in['instruments'][instrument]['refraction'].items():
                if key not in config_in['nights'][night]['refraction']:
                    config_in['nights'][night]['refraction'][key] = key_val

        #if 'master_out_method' not in config_in['nights'][night]:
        #    config_in['nights'][night]['master_out_method'] = None

        if config_in['nights'][night]['use_analytical_rvs'] and 'RV_semiamplitude' not in config_in['star']:
            print(" Missing RV_semiamplitude keyword for the star, the value will be computed from the RVs ")
            config_in['nights'][night]['use_analytical_rvs'] = False

    """ OLD approach to compute the RV of the planet, left here because it may be useful in the future
    try:

        _dict_star = {'mass': None, 'radius': None, 'gamma': None}
        for key in config_in['star']:
            _dict_star[key] = config_in['star'][key]

        config_in['system']['star'] = StarParameters(
            mass=_dict_star['mass'],
            radius=_dict_star['radius'])

        config_in['system']['common'] = {'degree': False, 'n_planets': 0, 'planets_list': []}
        for key in config_in['planets']['common']:
            config_in['system']['common'][key] = config_in['planets']['common'][key]

        for key in config_in['planets']:

            if key not in ['common']:
                config_in['system'][key] = PlanetParameters()
                config_in['system'][key].put_reference_epoch(config_in['system']['common']['Tref'])
                config_in['system'][key].put_RVparameters(
                    P=config_in['planets'][key]['P'],
                    K=config_in['planets'][key]['K'],
                    f=config_in['planets'][key]['Tc'],
                    e=config_in['planets'][key]['e'],
                    o=config_in['planets'][key]['o'],
                    degree=config_in['system']['common']['degree'])
                config_in['system'][key].put_RVplanet(config_in['planets'][key]['K_planet'])
                config_in['system'][key].put_star(config_in['system']['star'])
                config_in['system']['common']['n_planets'] += 1
                config_in['system']['common']['planets_list'].extend([key])
    except:
        pass
    """

    return config_in


def get_filelists(night_selected):
    """

    :param night_selected: usually the night_dict[night] dictionary from the main program
    :return:
    """

    """ List files are supposed to be in the same directory of the yaml file,
        NOT on the archive directory: in this way it is possible to try different
        combinations of nights and files without making a mess in the archive """

    files_list = np.atleast_1d(np.genfromtxt(night_selected['all'], dtype=str))

    try:
        files_transit_out = np.atleast_1d(np.genfromtxt(night_selected['out_transit'], dtype=str))
        files_transit_in = np.atleast_1d(np.genfromtxt(night_selected['in_transit'], dtype=str))
        files_transit_full = np.atleast_1d(np.genfromtxt(night_selected['full_transit'], dtype=str))
    except (FileNotFoundError, IOError):
        files_transit_out = None
        files_transit_in = None
        files_transit_full = None

    try:
        files_telluric = np.atleast_1d(np.genfromtxt(night_selected['telluric_list'], dtype=str))
    except (FileNotFoundError, IOError):
        files_telluric = None


    if night_selected['telluric'] is not None:
        files_star_telluric = np.atleast_1d(np.genfromtxt(night_selected['star_telluric'], dtype=str))
    else:
        files_star_telluric = None

    return files_list, files_transit_out, files_transit_in, files_transit_full, files_telluric, files_star_telluric


def from_config_get_nights(config_in):
    """
    This subroutine creates a shortcut to the night list
    :param config_in:
    :return: dictionary
    """
    return config_in['nights']


def from_config_get_instrument(config_in):
    """
    This subroutine creates a shortcut to the instrument list
    :param config_in:
    :return: dictionary
    """
    return config_in['instruments']


def from_config_refraction(config_in, night):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['nights'][night]['refraction']

def from_config_get_transmission_lightcurve(config_in):
    """
    This subroutine creates a shortcut to the transmission_lightcurve dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['transmission_lightcurve']
    except:
        return config_in['transmission']

def from_config_get_transmission(config_in):
    """
    This subroutine creates a shortcut to the transmission_lightcurve dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['transmission']
    except:
        return config_in['transmission_lightcurve']


def from_config_get_system(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['system']


def from_config_get_pipeline(config_in):
    """
    This subroutine creates a shortcut to the pipeline parameters dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['pipeline']


def from_config_get_planet(config_in):
    """
    This subroutine creates a shortcut to the planet dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['planet']


def from_config_get_star(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['star']


def from_config_get_clv_rm(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    return config_in['CLV_RM_correction']


def from_config_get_interstellar_lines(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['interstellar_lines']
    except:
        return None

def from_config_get_molecfit(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['molecfit']
    except:
        return None

def from_config_get_transmission_mcmc(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['transmission_mcmc']
    except:
        return None


def from_config_get_spectral_lines(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['spectral_lines']
    except:
        return None

def from_config_get_interactive_plots(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['interactive_plots']
    except:
        return False

def from_config_get_pca_parameters(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['pca_parameters']
    except:
        return {}


def from_config_get_fullspectrum_parameters(config_in):
    """
    This subroutine creates a shortcut to the system dictionary
    :param config_in:
    :return: dictionary
    """
    try:
        return config_in['full_spectrum']
    except:
        return {}